#!/usr/bin/env python3
"""
Efficiently extract a region from a FASTA file. When an .fai index of the file
is available, only the necessary parts of the file are read. If the index is
not available, the entire file is read into memory first. Create an index (.fai
file) with "samtools faidx".

The result is printed in FASTA format to standard output.

Regions are specified in the format "[rc:]name[:start-stop]".
If "start" and "stop" are omitted, the whole sequence is returned.
Coordinates are 1-based and both endpoints of the interval are included.
A region specification may be prefixed by 'rc:' in order to output the reverse
complement of the specified region. It must hold that start <= stop,
even when reverse complements are requested. If it does not hold, the output
sequence is empty.

Please be aware that samtools faidx uses only the part of the sequence name up to, but not including,
the first whitespace character. That is, if an entry in your FASTA file looks like this:

>seq1 this is a sequence

Then the identifier for this sequence is simply 'seq1'. For consistency, this
convention is also followed when the .fai file is not used.


Examples
--------

Extract chromosome 1 from a FASTA file named hg19.fa:

    fastaextract hg19.fa chr1

Extract the first 200 nucleotides from chromosome 1 in hg19.fa:

    fastaextract hg19.fa chr1:1-200

Extract the reverse complement of the bases 201 up to the end of chr1:

    fastaextract hg19.fa rc:chr1:201-

TODO
* create a .fai index on the fly instead of reading the full file into memory.
* check for duplicate names when no index used
"""


import os.path
import sys
import mmap

from .. import HelpfulArgumentParser
from ..io.fasta import (FastaReader, NonIndexedFasta, IndexedFasta, FastaWriter,
	FastaIndexMissing)
from ..io.xopen import xopen
from ..dna import reverse_complement
from ..region import Region

__author__ = "Marcel Martin"


def main():
	if sys.version < '2.7':
		print("Sorry, Python version >= 2.7 required!", file=sys.stderr)
		sys.exit(1)
	parser = HelpfulArgumentParser(description=__doc__)
	parser.add_argument("--width", "-w", type=int, default=80,
		help="Characters per line in output FASTA (default: %(default)s). "
			"Set to 0 to disallow line breaks entirely.")
	parser.add_argument("fasta", metavar="FASTA", help="The FASTA file")
	parser.add_argument("region", metavar="REGION", nargs='+')
	args = parser.parse_args()
	if args.width == 0:
		args.width = None

	try:
		fasta = IndexedFasta(args.fasta)
	except FastaIndexMissing:
		if os.path.getsize(args.fasta) > 1024 ** 3:  # 1 GiB
			print("ERROR: The file is very large and no index exists, "
				"please create an index with 'samtools faidx'.", file=sys.stderr)
			sys.exit(1)
		fasta = NonIndexedFasta(args.fasta)

	writer = FastaWriter(sys.stdout, line_length=args.width)
	regions = [ Region(s) for s in args.region ]
	for region in regions:
		sequence = fasta[region.reference][region.start:region.stop]
		if region.is_reverse_complement:
			sequence = reverse_complement(sequence)
		if sys.version > '3':
			sequence = sequence.decode()
		writer.write(str(region), sequence)


if __name__ == '__main__':
	main()



"""
Classes that help in working with supplementary alignments.
"""
from collections import namedtuple
from .cigar import Cigar

class SupplementaryAlignment(namedtuple("SupplementaryAlignment",
	['reference_name', 'pos', 'strand', 'cigar', 'mapping_quality', 'edit_distance'])):
	@property
	def query_start(self):
		return self.cigar.clipping_left

	@property
	def query_end(self):
		return self.cigar.clipping_left + self.cigar.query_length(count_clipped=None)

	@property
	def query_length(self):
		return self.cigar.query_length(count_clipped=None)

	@property
	def reference_end(self):
		return self.pos + self.cigar.reference_length()


def parse_supplementary(sa):
	"""
	Parse supplementary alignments given by the SA:Z: tag in a SAM record.

	Return a list of SupplementaryAlignment objects.
	"""
	fields = sa.split(';')
	assert fields[-1] == ''  # sa must end with a ';'
	alignments = []
	for field in fields[:-1]:
		ref, pos, strand, cig, mapq, edit_dist = field.split(',')
		pos = int(pos) - 1
		cig = Cigar(cig)

		# All information in the BAM file is relative to the reference.
		# Since we are interested in information relative to the read,
		# the CIGAR string needs to be reversed when the strand is '-'.
		if strand == '-':
			cig = cig[::-1]
		mapq = int(mapq)
		edit_dist = int(edit_dist)
		assert strand in '+-'
		a = SupplementaryAlignment(ref, pos, strand, cig, mapq, edit_dist)
		alignments.append(a)
	return alignments


class AlignedRead:
	"""
	An AlignedRead describes all alignments of a read that appears in a
	SAM/BAM file. This collects all the supplementary alignments for the read in
	one place and gives a read-centric view of the alignments, instead of a
	reference-centric one.
	"""
	def __init__(self, read, reference_name):
		"""
		read -- an AlignedSegment. The SA tag of the read is parsed.
		"""
		#assert not read.is_secondary, "Read should not be secondary"
		#assert not read.is_supplementary, "Read should not be supplementary"
		self.alignments = self._extract_supplementary_alignments(read, reference_name)
		self.query_name = read.query_name
		self._primary_read = read  # TODO not necessarily correct since assertions were removed above
		self._length = self.alignments[0].cigar.query_length(count_clipped='hard')
		#assert self._length == Cigar(read.cigar).query_length(count_clipped='soft'), "Shouldn't be hard-clipped"

	def __len__(self):
		"""
		Number of bases in this read.
		"""
		return self._length

	def __iter__(self):
		yield from self.alignments

	def aligned_bases(self):
		"""
		Return how many bases are aligned, considering all supplementary
		alignments. Overlapping supplementary alignments are allowed: Each base
		is counted at most once!
		"""
		events = []
		for alignment in self.alignments:
			events.append((alignment.query_start, 'start', None))
			events.append((alignment.query_end, 'stop', alignment))

		depth = 0  # number of observed 'start' events
		bases = 0  # number of covered bases
		last_qstart = None
		for qpos, what, alignment in sorted(events, key=lambda x: x[0]):
			if what == 'start':
				if depth == 0:
					last_qstart = qpos
				depth += 1
			elif what == 'stop':
				depth -= 1
				if depth == 0:
					# interval (last_qstart, qpos) was covered
					bases += qpos - last_qstart
		return bases

	def _extract_supplementary_alignments(self, aligned_segment, reference_name):
		"""
		Given a single entry in a BAM file (an AlignedSegment) that potentially
		has supplementary alignments specified by the SA tag, return a list of
		SupplementaryAlignment objects. The list includes at least the aligned
		segment itself.
		"""
		tags = dict(aligned_segment.tags)
		cig = Cigar(aligned_segment.cigar)
		if aligned_segment.is_reverse:
			strand = '-'
			cig = cig[::-1]
		else:
			strand = '+'
		alignments = [SupplementaryAlignment(reference_name, aligned_segment.pos,
			strand, cig, aligned_segment.mapq, tags['NM'])]
		if 'SA' in tags:
			alignments.extend(parse_supplementary(tags['SA']))
		if __debug__:
			l = alignments[0].cigar.query_length(count_clipped='hard')
			for alignment in alignments[1:]:
				assert alignment.cigar.query_length(count_clipped='hard') == l
		return alignments



#!/usr/bin/env python3
"""
Print lots of statistics about one or more FASTA or FASTQ files.

TODO
- computation of contig N50 is relatively slow
"""
import sys
import os
import subprocess
from collections import Counter
from .. import HelpfulArgumentParser, SequenceReader, IndexedFasta
from ..dna import n_intervals, intervals_complement
from ..math import n50, frequency_median, frequency_n50

__author__ = "Marcel Martin"


def byte_frequencies(s):
	return Counter(s)


try:
	from sqt._helpers import byte_frequencies
except:
	pass


def print_statistics(lengths, contig_lengths, shortest=None, longest=None, genome_size=None, character_frequencies=None):
	"""
	lengths -- a dictionary that maps length to frequency (Counter object)
	"""
	n = sum(lengths.values())
	print('No. of sequences: {:11,}'.format(n))
	if n == 0:
		return

	total = sum(length * count for length, count in lengths.items())
	print('Total length:   {:13,}'.format(total))

	min_length, max_length = min(lengths), max(lengths)
	if shortest:
		print('Minimum length: {:13,} in entry "{}"'.format(min_length, shortest))
	else:
		print('Minimum length: {:13,}'.format(min_length))
	if longest:
		print('Maximum length: {:13,} in entry "{}"'.format(max_length, longest))
	else:
		print('Maximum length: {:13,}'.format(max_length))

	print('Average length: {:16.2f}'.format(total / n))
	print('Median length:  {:13,}'.format(frequency_median(lengths)))
	print('Scaffold N50:   {:13,}'.format(frequency_n50(lengths)))
	if genome_size:
		print('Scaffold NG50:  {:13,}'.format(frequency_n50(lengths, genome_size=genome_size)))

	if contig_lengths:
		lengths = contig_lengths
		min_length, max_length = min(lengths), max(lengths)
		print()
		n_contigs = sum(lengths.values())
		print('Number of contigs:     {:13,}'.format(n_contigs))
		total_c = sum(length * count for length, count in lengths.items())
		print('Total contig length:   {:13,}'.format(total_c))
		print('Minimum contig length: {:13,}'.format(min_length))
		print('Maximum contig length: {:13,}'.format(max_length))
		print('Average contig length: {:16.2f}'.format(total_c / n_contigs))
		print('Median contig length:  {:13,}'.format(frequency_median(lengths)))
		print('Contig N50:            {:13,}'.format(frequency_n50(lengths)))

	if character_frequencies:
		print()
		print("Character distribution (<char> <count> <percentage>):")
		assert total == sum(character_frequencies.values())
		acgt = 0
		gc = 0
		for upper, lower in (b'Aa', b'Cc', b'Gg', b'Tt'):
			freq = character_frequencies[upper] + character_frequencies[lower]
			if upper in b'GC':
				gc += freq
			print(chr(upper), '    {:14,} {:6.1%}'.format(freq, freq / total))
			acgt += freq
		other = sum(character_frequencies.values()) - acgt
		print('other {:14,} {:6.2%}'.format(other, other / total))
		print('ACGT  {:14,} {:6.2%}'.format(acgt, acgt / total))
		print('GC    {:14,} {:6.2%} (of ACGT)'.format(gc, gc / (total - other)))


def filter_short_intervals(intervals, minimum_length):
	for start, stop in intervals:
		if stop - start >= minimum_length:
			yield (start, stop)


def fasta_fastq_iter(path):
	with SequenceReader(path, 'rb') as reader:
		for record in reader:
			seq = record.sequence #.upper()
			yield (record.name, len(seq), seq)


def indexed_fasta_iter(path):
	with IndexedFasta(path) as f:
		for index_entry in f.index.values():
			yield (index_entry.name, index_entry.length, None)


def stats(path, tolerable_gapsize, detailed):
	"""
	Determine scaffold lengths, contig lengths and character frequencies.

	Return a tuple (scaffold_lengths, contig_lengths, character_frequencies).
	"""
	scaffold_lengths = Counter()
	contig_lengths = Counter()
	nucleotides = Counter()  # nucleotide frequencies
	shortest = None
	longest = None
	min_length = float('+inf')
	max_length = -1
	if not detailed and os.path.exists(path + '.fai'):
		it = indexed_fasta_iter(path)
	else:
		it = fasta_fastq_iter(path)
	for (name, length, sequence) in it:
		scaffold_lengths[length] += 1
		if length < min_length:
			min_length = length
			shortest = name
		if length > max_length:
			max_length = length
			longest = name
		if detailed and sequence is not None:
			nucleotides += byte_frequencies(sequence)
			intervals = intervals_complement(
				filter_short_intervals(n_intervals(sequence, ord(b'N')), tolerable_gapsize), length)
			for start, stop in intervals:
				contig_lengths[stop - start] += 1

	return scaffold_lengths, contig_lengths, shortest, longest, nucleotides


def get_argument_parser():
	parser = HelpfulArgumentParser(description=__doc__)
	add = parser.add_argument
	add('--detailed', '-d', default=False, action='store_true',
		help='Print information about the sequences themselves, '
			'such as the character distribution and contig N50.')
	add('--genome-size', '-g', type=int, default=None,
		help='Estimated genome size. If given, also NG50 in addition to N50 is computed.')
	add('--tolerable-gapsize', '-t', type=int, default=10,
		help='A stretch of at most this many "N"s is not counted as a gap '
		'separating contigs.')
	add('fastaq', nargs='+', metavar='FASTA/FASTQ',
		help='Input FASTA or FASTQ file(s) (may be gzipped).')
	return parser


def main():
	parser = get_argument_parser()
	args = parser.parse_args()

	overall_frequencies = Counter()
	overall_lengths = Counter()
	if not args.detailed:
		character_frequencies = None
		contig_lengths = None
	for path in args.fastaq:
		print("## File:", path)
		scaffold_lengths, contig_lengths, shortest, longest, character_frequencies = \
			stats(path, args.tolerable_gapsize, detailed=args.detailed)
		overall_frequencies += character_frequencies
		print_statistics(scaffold_lengths, contig_lengths, shortest, longest, args.genome_size, character_frequencies)
		overall_lengths += scaffold_lengths
	if len(args.fastaq) > 1:
		print("## Summary of", len(args.fastaq), "files")
		print_statistics(overall_lengths, None, None, None, args.genome_size, overall_frequencies if args.detailed else None)


if __name__ == '__main__':
	main()



from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from .args import HelpfulArgumentParser
from .io.fasta import (
	SequenceReader, FastaReader, FastqReader, FastaWriter, FastqWriter,
	IndexedFasta, guess_quality_base )
from .io.gtf import GtfReader
from cutadapt.xopen import xopen
from .cigar import Cigar




"""
FASTA and FASTQ I/O

- read and write FASTA and FASTQ files
- random access to indexed FASTA files that have a .fai index
- guess FASTQ quality encoding
- obtain header information (barcode, lane, flowcell id) from a FASTQ file
"""

from collections import namedtuple, OrderedDict, Counter
from itertools import islice
import mmap
import sys
import re
import gzip
from os.path import splitext
from .xopen import xopen
from ..compat import force_str

__author__ = "Marcel Martin"

# use this constant with bytes.translate to convert encoded quality values from
# ascii(phred_quality + 64) to ascii(phred_quality + 33)
TRANSLATE_64_TO_33 = ''.join(chr(max(c - 64 + 33, 0)) for c in range(256))


def _shorten(s, n=20):
	"""Shorten string s to at most n characters, appending "..." if necessary."""
	if s is None:
		return None
	if len(s) > n:
		ellipsis = '...' if isinstance(s, str) else b'...'
		s = s[:n-3] + ellipsis
	return s


def _quality_to_ascii(qualities, base=33):
	"""
	Convert a list containing qualities given as integer to a string of
	ASCII-encoded qualities.

	base -- ASCII code of quality zero (sensible values are 33 and 64).

	>>> _quality_to_ascii([17, 4, 29, 18])
	'2%>3'
	"""
	qualities = ''.join(chr(q+base) for q in qualities)
	return qualities


class UnknownFileType(Exception):
	"""
	Raised when SequenceReader could not autodetect the file type.
	"""
	pass


class FastaIndexMissing(Exception):
	pass


class FastaWriter:
	"""
	Write FASTA-formatted sequences to a file-like object.
	"""
	def __init__(self, file, line_length=80):
		"""
		If line_length is not None, the lines will
		be wrapped after line_length characters.
		"""
		self.line_length = line_length if line_length != 0 else None
		if isinstance(file, str):
			file = xopen(file, "w")
		self._file = file

	def write(self, name_or_seq, sequence=None):
		"""Write an entry to the the FASTA file.

		If only one parameter (name_or_seq) is given, it must have
		attributes .name and .sequence, which are then used.
		Otherwise, the first parameter must be the name and the second
		the sequence.

		The effect is that you can write this:
		fr.write("name", "ACCAT")
		or
		fr.write(Sequence("name", "ACCAT"))
		"""
		if sequence is None:
			name = name_or_seq.name
			sequence = name_or_seq.sequence
		else:
			name = name_or_seq
		sequence = force_str(sequence)
		if self.line_length is not None:
			print('>{}'.format(name), file=self._file)
			for i in range(0, len(sequence), self.line_length):
				print(sequence[i:i+self.line_length], file=self._file)
		else:
			print('>{}'.format(name), sequence, file=self._file, sep='\n')

	def close(self):
		self._file.close()

	def __enter__(self):
		if self._file.closed:
			raise ValueError("I/O operation on closed file")
		return self

	def __exit__(self, *args):
		self.close()


class FastqWriter:
	"""
	Write sequences with qualities in FASTQ format.

	FASTQ files are formatted like this:
	@read name
	SEQUENCE
	+
	QUALITIS
	"""
	_close_on_exit = False

	def __init__(self, file, twoheaders=False):
		"""
		If twoheaders is set, then the read name will be repeated after
		the plus sign (which is redundant and therefore not
		recommended).
		"""
		self.twoheaders = twoheaders
		if isinstance(file, str):
			file = xopen(file, "w")
			self._close_on_exit = True
		self._file = file

	def write(self, name_or_seq, sequence=None, qualities=None):
		"""Write an entry to the the FASTQ file.

		If only one parameter (name_or_seq) is given, it must have
		attributes .name, .sequence and .qualities, which are then
		used. Otherwise, all three parameters must be given and
		name_or_seq must be the name of the sequence.

		The effect is that you can write this:
		fq.write("name", "ACCAT", "#!!&B")
		or
		fq.write(Sequence("name", "ACCAT", "#!!&B"))
		"""
		if sequence is None:
			name = name_or_seq.name
			sequence = name_or_seq.sequence
			qualities = name_or_seq.qualities
		else:
			name = name_or_seq
		if self.twoheaders:
			two = name
		else:
			two = ''
		print("@{0}\n{1}\n+{2}\n{3}".format(name, force_str(sequence), two, force_str(qualities)), file=self._file)

	def close(self):
		if self._close_on_exit:
			self._file.close()

	def __enter__(self):
		if self._file.closed:
			raise ValueError("I/O operation on closed file")
		return self

	def __exit__(self, *args):
		self.close()


IndexEntry = namedtuple("IndexEntry", "name length offset nucleotides_per_line bytes_per_line")


def _read_fasta_index(path):
	"""
	Return a dictionary that maps sequence names to IndexEntry tuples.

	The sequence name is the string before the first space character in the FASTA comment header.

	The offsets are converted to Python 0-based coordinates!
	"""
	index = OrderedDict()
	try:
		fai = open(path)
	except FileNotFoundError as e:
		raise FastaIndexMissing("Could not find the FASTA index file named '{}'.".format(path))
	with fai as f:
		for line in f:
			name, length, offset, nucleotides_per_line, bytes_per_line = line.split('\t')
			indexed_name = name.split(' ', maxsplit=1)[0]
			length = int(length)
			offset = int(offset)
			nucleotides_per_line = int(nucleotides_per_line)
			bytes_per_line = int(bytes_per_line)
			entry = IndexEntry(name, length, offset, nucleotides_per_line, bytes_per_line)
			index[indexed_name] = entry
	return index


class IndexedSequence:
	"""A single sequence in an indexed FASTA file"""

	def __init__(self, mapped, indexentry):
		self.mapped = mapped
		self.indexentry = indexentry

	def __getitem__(self, key):
		if type(key) is int:
			key = slice(key, key+1)
		assert key.step is None

		return self.read(key.start, key.stop)

	def __len__(self):
		"""Return length of sequence"""
		return self.indexentry.length

	def read(self, start=0, stop=None):
		"""
		DEPRECATED, use slice notation via sequence[5:20] instead.

		Read and return a substring of a specific entry of the FASTA file
		"""
		indexinfo = self.indexentry
		if stop is None:
			stop = indexinfo.length
		if start is None:
			start = 0
		start = max(0, start)
		stop = min(indexinfo.length, stop)
		if stop <= start:
			return b''

		def nucleotide_to_byte_offset(i):
			return indexinfo.offset + i // indexinfo.nucleotides_per_line * indexinfo.bytes_per_line + i % indexinfo.nucleotides_per_line

		byte_start = nucleotide_to_byte_offset(start)
		byte_stop = nucleotide_to_byte_offset(stop - 1) + 1

		# these indices are relative to this sequence
		start_line = start // indexinfo.nucleotides_per_line
		stop_line = (stop - 1) // indexinfo.nucleotides_per_line + 1

		assert start_line < stop_line

		if start_line + 1 == stop_line:
			# requested substring is within one line
			return self.mapped[byte_start:byte_stop]

		# otherwise, collect lines or line fragments
		collected = []
		# offset of end of start_line
		byte_line_stop = indexinfo.offset + start_line * indexinfo.bytes_per_line + indexinfo.nucleotides_per_line
		assert byte_start < byte_line_stop
		collected.append(self.mapped[byte_start:byte_line_stop])

		# all lines between the start_- and stop_line
		for line in range(start_line + 1, stop_line - 1):
			byte_line_start = indexinfo.offset + line * indexinfo.bytes_per_line
			assert byte_line_start < byte_line_start + indexinfo.nucleotides_per_line
			collected.append(self.mapped[byte_line_start:byte_line_start+indexinfo.nucleotides_per_line])

		# last line
		byte_line_start = indexinfo.offset + (stop_line - 1) * indexinfo.bytes_per_line
		assert byte_line_start < byte_stop
		collected.append(self.mapped[byte_line_start:byte_stop])

		sequence = b''.join(collected)
		assert len(sequence) == stop - start
		return sequence


class IndexedFasta:
	"""
	Efficient access to FASTA files that have been indexed by 'samtools faidx'.

	Example:

	with IndexedFasta(path) as fasta:
		print(fasta['chr1'][50:100])

	In Python 3, the returned sequence is of type bytes.
	"""

	def __init__(self, path):
		self.index = _read_fasta_index(path + '.fai')
		self._file = open(path)
		self.mapped = mmap.mmap(self._file.fileno(), 0, flags=mmap.MAP_PRIVATE)  # 0: entire file

	def length(self, name):
		"""
		DEPRECATED, use len(indexedfasta[name]) instead

		Return length of sequence 'name'.
		"""
		return self.index[name].length

	def get(self, name):
		"""
		DEPRECATED, use indexing with indexed_fasta[name] instead.

		Retrieve an entry of the FASTA file as an IndexedSequence object.
		This is a light-weight operation: The actual work happens when you
		access the returned object.
		"""
		return IndexedSequence(self.mapped, self.index[name])

	def __getitem__(self, name):
		"""
		Retrieve an entry of the FASTA file as an IndexedSequence object.
		This is a light-weight operation: The actual work happens when you
		access the returned object.
		"""
		if not isinstance(name, str):
			raise TypeError("The key must be a str.")
		return IndexedSequence(self.mapped, self.index[name])

	def __len__(self):
		return len(self.index)

	def __contains__(self, name):
		return name in self.index

	def close(self):
		self.mapped.close()
		self._file.close()

	def __enter__(self):
		if self._file.closed:
			raise ValueError("I/O operation on closed file")
		return self

	def __exit__(self, *args):
		self.close()


def NonIndexedFasta(path):
	"""
	Efficient access to FASTA files that have not been indexed with 'samtools faidx'.

	In contrast to IndexedFasta above, this implementation reads the entire
	file into memory (into a dictionary) to provide efficient access.

	Return a dict that maps sequence names to sequences.

	In Python 3, the sequences are of type bytes.
	"""
	with FastaReader(path, wholefile=True, mode='rb') as fr:
		sequences = dict((seq.name.split(' ', maxsplit=1)[0], seq.sequence) for seq in fr)
	return sequences


class Sequence:
	"""qualities is a string and it contains the qualities encoded as ascii(qual+33)."""

	def __init__(self, name, sequence, qualities=None):
		"""Set qualities to None if there are no quality values"""
		self.name = name
		self.sequence = sequence
		self.qualities = qualities

	def __getitem__(self, key):
		"""slicing"""
		return self.__class__(self.name, self.sequence[key], self.qualities[key] if self.qualities is not None else None)

	def __repr__(self):
		qstr = ''
		if self.qualities is not None:
			qstr = '\', qualities={0!r}'.format(_shorten(self.qualities))
		return 'Sequence(name={0!r}, sequence={1!r}{2})'.format(_shorten(self.name), _shorten(self.sequence), qstr)

	def __len__(self):
		return len(self.sequence)

	def __eq__(self, other):
		return self.name == other.name and \
			self.sequence == other.sequence and \
			self.qualities == other.qualities

	def __ne__(self, other):
		return not self.__eq__(other)

	def __hash__(self):
		return hash((self.name, self.sequence, self.qualities))


class FormatError(Exception):
	"""
	Raised when an input file (FASTA or FASTQ) is malformatted.
	"""
	pass


class FileWithPrependedLine(object):
	"""
	A file-like object that allows to "prepend" a single
	line to an already opened file. That is, further
	reads on the file will return the provided line and
	only then the actual content. This is needed to solve
	the problem of autodetecting input from a stream:
	As soon as the first line has been read, we know
	the file type, but also that line is "gone" and
	unavailable for further processing.
	"""
	def __init__(self, file, line):
		"""
		file is an already opened file-like object.
		line is a single string (newline will be appended if not included)
		"""
		if not line.endswith('\n'):
			line += '\n'
		self.first_line = line
		self._file = file

	def __iter__(self):
		yield self.first_line
		for line in self._file:
			yield line

	def close(self):
		self._file.close()

	@property
	def closed(self):
		return self._file.closed



def SequenceReader(file, mode='rt', colorspace=False, fileformat=None):
	"""
	Reader for FASTA and FASTQ files that autodetects the file format.
	Returns either an instance of FastaReader or of FastqReader,
	depending on file type.

	The autodetection can be skipped by setting fileformat to the string
	'fasta' or 'fastq'

	file is a filename or a file-like object.
	If file is a filename, then .gz files are supported.
	If the file name is available, the file type is detected
	by looking at the file name.
	If the file name is not available (for example, reading
	from standard input), then the file is read and the file
	type determined from the content.
	"""
	if fileformat is not None:
		fileformat = fileformat.lower()
		if fileformat == 'fasta':
			return FastaReader(file, mode=mode)
		elif fileformat == 'fastq':
			return FastqReader(file, mode=mode)
		else:
			raise UnknownFileType("File format {0} is unknown (expected 'fasta' or 'fastq').".format(fileformat))

	name = None
	if file == "-":
		file = sys.stdin
	elif isinstance(file, str):
		name = file
	elif hasattr(file, "name"):
		name = file.name
	if name is not None:
		if name.endswith('.gz'):
			name = name[:-3]
		name, ext = splitext(name)
		ext = ext.lower()
		if ext in ['.fasta', '.fa', '.fna', '.csfasta', '.csfa']:
			return FastaReader(file, mode=mode)
		elif ext in ['.fastq', '.fq'] or (ext == '.txt' and name.endswith('_sequence')):
			return FastqReader(file, colorspace=colorspace, mode=mode)
		else:
			raise UnknownFileType("Could not determine whether this is FASTA or FASTQ: file name extension {0} not recognized".format(ext))

	# No name available.
	# Assume that 'file' is an open file
	# and autodetect its type by reading from it.
	for line in file:
		if line.startswith('#'):
			# Skip comment lines (needed for csfasta)
			continue
		if line.startswith('>'):
			return FastaReader(FileWithPrependedLine(file, line), mode=mode)
		if line.startswith('@'):
			return FastqReader(FileWithPrependedLine(file, line), colorspace=colorspace, mode=mode)
	raise UnknownFileType("File is neither FASTQ nor FASTA.")



class FastaReader(object):
	"""
	Reader for FASTA files.
	"""
	def __init__(self, file, mode='rt', wholefile=False, keep_linebreaks=False, binary=None):
		"""
		file is a filename or a file-like object.
		If file is a filename, then .gz files are supported.
		If wholefile is True, then it is ok to read the entire file
		into memory. This is faster when there are many newlines in
		the file, but may obviously need a lot of memory.
		keep_linebreaks -- whether to keep the newline characters in the sequence

		binary -- DEPRECATED, use mode='rb' instead. In Python 3, when this is
		set, the returned Sequence objects will contain bytes objects for the
		sequence. The names will still be strings. If file was given as a
		file-like object, it must have been opened in binary mode.

		mode -- When set to 'rt', returned Sequence objects will contain bytes
		objects for the sequence (relevant in Python 3 only).
		"""
		if binary is not None:
			mode = 'rb' if binary else 'rt'
		if mode not in ('rt', 'rb'):
			raise ValueError("mode must be either 'rt' or 'rb'")
		if isinstance(file, str):
			file = xopen(file, mode)
		self._file = file
		self.binary = binary
		self.wholefile = wholefile
		self.keep_linebreaks = keep_linebreaks
		self.mode = mode
		self.format = 'fasta'
		assert not (wholefile and keep_linebreaks), "not supported"

	def __iter__(self):
		"""
		Yield Sequence objects.
		The qualities attribute is always None.
		"""
		return self._wholefile_iter() if self.wholefile else self._streaming_iter()

	def _streaming_iter(self):
		"""
		Read entry from the file (single entry at a time).
		"""
		name = None
		if self.mode == 'rt':
			seq_constructor = str
			delim = '\n'
			startchar = '>'
			seq_type = lambda s: s
			name_transform = lambda name: name
		else:
			seq_constructor = bytearray
			seq_type = lambda s: bytes(s)
			delim = b'\n'
			startchar = ord('>')
			name_transform = lambda name: name.decode('utf-8')

		seq = seq_constructor()
		for line in self._file:
			# strip() should also take care of DOS line breaks
			line = line.strip()
			if line and line[0] == startchar:
				if name is not None:
					assert self.keep_linebreaks or seq.find(delim) == -1
					name = name_transform(name)
					seq = seq_type(seq)
					yield Sequence(name, seq, None)
				name = line[1:]
				seq = seq_constructor()
			else:
				seq += line
				if self.keep_linebreaks:
					seq += delim
		if name is not None:
			assert self.keep_linebreaks or seq.find(delim) == -1
			name = name_transform(name)
			seq = seq_type(seq)
			yield Sequence(name, seq, None)

	def _wholefile_iter(self):
		"""
		This reads in the entire file at once, but is faster than the above code when there are lots of newlines.
		The idea comes from the TAMO package (http://fraenkel.mit.edu/TAMO/), module TAMO.seq.Fasta by
		David Benjamin Gordon.
		"""
		wholefile = self._file.read()
		if self.mode == 'rb':
			cr = b'\r'
			linebreak = b'\n'
			delim = b'\n>'
			greater = b'>'
			to_str = lambda s: s.decode()
		else:
			cr = '\r'
			linebreak = '\n'
			delim = '\n>'
			greater = '>'
			to_str = lambda s: s
		assert cr not in wholefile, "Sorry, currently don't know how to deal with files that contain \\r linebreaks"
		assert len(wholefile) == 0 or wholefile[0:1] == greater, "FASTA file must start with '>'"
		parts = wholefile.split(delim)
		# first part has '>' in front
		parts[0] = parts[0][1:]
		for part in parts:
			lines = part.split(linebreak, 1)
			name = to_str(lines[0])
			sequence = lines[1].replace(linebreak, bytes())
			yield Sequence(name=name, sequence=sequence, qualities=None)

	def close(self):
		self._file.close()

	def __enter__(self):
		if self._file.closed:
			raise ValueError("I/O operation on closed FastaReader")
		return self

	def __exit__(self, *args):
		self._file.close()


class FastqReader(object):
	"""
	Reader for FASTQ files. Does not support multi-line FASTQ files.
	"""
	def __init__(self, file, mode='rt', colorspace=False):
		"""
		file is a filename or a file-like object.
		If file is a filename, then the file is opened with xopen().

		colorspace -- Usually (when this is False), there must be n characters in the sequence and
		 n quality values. When this is True, there must be n+1 characters in the sequence and n quality values.

		mode -- For Python 3, set this to 'rb' to get Sequence objects in which
		 both the sequence and the qualities field have type bytes.
		"""
		if mode not in ('rt', 'rb'):
			raise ValueError("mode must be either 'rt' or 'rb'")
		if isinstance(file, str):
			file = xopen(file, mode)
		self._file = file
		self.colorspace = colorspace
		self.twoheaders = False
		self.mode = mode
		self.format = 'fastq'

	def __iter__(self):
		"""
		Yield Sequence objects.
		"""
		if self.mode == 'rt':
			AT = '@'
			PLUS = '+'
			STRIP = '\n\r'
			name_transform = lambda name: name
		else:
			AT = b'@'
			PLUS = b'+'
			STRIP = b'\n\r'
			name_transform = lambda name: name.decode('utf-8')
		lengthdiff = 1 if self.colorspace else 0
		for i, line in enumerate(self._file):
			if i % 4 == 0:
				if not line.startswith(AT):
					raise FormatError("at line {0}, expected a line starting with '@'".format(i+1))
				name = name_transform(line.strip()[1:])
			elif i % 4 == 1:
				sequence = line.strip()
			elif i % 4 == 2:
				line = line.strip()
				if not line.startswith(PLUS):
					raise FormatError("at line {0}, expected a line starting with '+'".format(i+1))
				if len(line) > 1:
					self.twoheaders = True
					if not name_transform(line[1:]) == name:
						raise FormatError(
							"At line {0}: Two sequence descriptions are given in "
							"the FASTQ file, but they don't match "
							"('{1}' != '{2}')".format(i+1, name, line.rstrip()[1:]))
			elif i % 4 == 3:
				qualities = line.rstrip(STRIP)
				if len(qualities) + lengthdiff != len(sequence):
					raise ValueError("Length of quality sequence and length of read do not match (%d+%d!=%d)" % (len(qualities), lengthdiff, len(sequence)))
				yield Sequence(name, sequence, qualities)

	def __enter__(self):
		if self._file is None:
			raise ValueError("I/O operation on closed FastqReader")
		return self

	def __exit__(self, *args):
		self._file.close()


class FastaQualReader(object):
	"""
	Reader for reads that are stored in .(CS)FASTA and .QUAL files.
	"""
	def __init__(self, fastafile, qualfile, colorspace=False):
		"""
		fastafile and qualfile are filenames file-like objects.
		If file is a filename, then .gz files are supported.

		colorspace -- Usually (when this is False), there must be n characters in the sequence and
		n quality values. When this is True, there must be n+1 characters in the sequence and n quality values.
		"""
		self.fastareader = FastaReader(fastafile)
		self.qualreader = FastaReader(qualfile, keep_linebreaks=True)
		self.colorspace = colorspace

	def __iter__(self):
		"""
		Yield Sequence objects.
		"""
		lengthdiff = 1 if self.colorspace else 0
		for fastaread, qualread in zip(self.fastareader, self.qualreader):
			qualities = _quality_to_ascii(list(map(int, qualread.sequence.split())))
			assert fastaread.name == qualread.name
			if len(qualities) + lengthdiff != len(fastaread.sequence):
				raise ValueError("Length of quality sequence and length of read do not match (%d+%d!=%d)" % (
					len(qualities), lengthdiff, len(fastaread.sequence)))
			yield Sequence(fastaread.name, fastaread.sequence, qualities)

	def __enter__(self):
		if self.fastafile is None:
			raise ValueError("I/O operation on closed FastaQualReader")
		return self

	def __exit__(self, *args):
		self.fastareader.close()
		self.qualreader.close()


def byte_frequencies(s):
	return Counter(s)


def guess_quality_base(path, limit=10000):
	"""
	Guess quality encoding (Phred33/64) of a FASTQ file.

	Return a tuple (freqs, guess) where freqs are the character frequencies
	encountered in the first limit records and guess is one of 33, 64, None,
	indicating which quality encoding this file probably has (None indicates
	unknown).
	"""
	freqs = Counter()
	with FastqReader(path, mode='rb') as fqr:
		for record in islice(fqr, 0, limit):
			freqs += byte_frequencies(record.qualities)
	if min(freqs) - 64 < -10:
		guess = 33
	elif max(freqs) - 33 > 60:
		guess = 64
	else:
		guess = None
	return freqs, guess


FastqInfo = namedtuple('FastqInfo', ['instrument', 'run', 'flowcell', 'lane', 'barcode'])

def fastq_header(path):
	"""
	Inspect a FASTQ file and return a FastqInfo object. If a particular piece of
	information is unknown, the corresponding attribute of FastqInfo is set to
	None.

	This will try to auto-detect different types of Illumina headers:
	@HWI-ST552_0:4:1101:1179:1939#0/1
	@HWI_ST139:8:1:1202:1874#GATCAG/1
	@HWI-ST344:204:D14G8ACXX:8:1101:1638:2116 1:N:0:CGATGT
	@MISEQ:56:000000000-A4YM7:1:1101:15071:2257 1:N:0:CTTGTA
	@FCD20MKACXX:8:1101:1215:2155#TCGTAAGC/1

	The format of a FASTQ header starting with CASAVA 1.8 is:

	<instrument-name>:<run ID>:<flowcell ID>:<lane-number>:<tile-number>:<x-pos>:<y-pos>
	<read number>:<is filtered>:<control number>:<barcode sequence>
	"""
	try:
		#
		line = path.readline()
	except:
		if path.endswith('.gz'):
			openfunc = gzip.open
		else:
			openfunc = open
		with openfunc(path, 'rt') as f:
			line = f.readline()
	line = line.rstrip()
	assert line.startswith('@')
	line = line[1:]
	# ignore comment field for now
	header = line.split(' ', maxsplit=1)
	read_name = header[0]
	comment = header[1] if len(header) == 2 else None
	fields = read_name.split(':')
	if len(fields) == 7:
		# probably new CASAVA 1.8 format
		instrument = fields[0]
		run_id = int(fields[1])
		flowcell = fields[2]
		lane = int(fields[3])
		barcode = None
		if comment:
			comment_fields = comment.split(':')
			if len(comment_fields) == 4:
				barcode = comment_fields[3]
	elif len(fields) == 5:
		if '#' in fields[4]:
			f = fields[4].split('#', maxsplit=1)[1]
			barcode = f.split('/', maxsplit=1)[0]
		else:
			barcode = None
		run_id = None
		lane = int(fields[1])
		if fields[0].endswith('XX'):
			instrument, flowcell = None, fields[0]
		else:
			instrument, flowcell = fields[0], None
	else:
		raise ValueError("FASTQ header format not recognized")
	if barcode and not re.match('[acgtnACGTN]+$', barcode):
		barcode = None

	return FastqInfo(instrument=instrument, run=run_id, flowcell=flowcell, lane=lane, barcode=barcode)


try:
	from .._helpers import byte_frequencies
except:
	pass



#!/usr/bin/env python3
"""
SeQuencing Tools -- command-line tools for working with sequencing data
"""
__author__ = "Marcel Martin"

import logging
import importlib
from . import HelpfulArgumentParser

from . import __version__
from .commands import fastxmod

logger = logging.getLogger(__name__)


# List of all subcommands. A module of the given name must exist and define
# add_arguments() and main() functions. Documentation is taken from the first
# line of the module’s docstring.
COMMANDS = [
	'align',
	'bam2fastq',
	'fastxmod',
	'qgramfreq',
	'chars',
	'fastagrep',
	'readcov',
	'randomseq',
	'samsetop',
	'bameof',
	'readlenhisto',
	'cutvect',
]

def main():
	logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')
	parser = HelpfulArgumentParser(description=__doc__, prog='sqt')
	parser.add_argument('--version', action='version', version='%(prog)s ' + __version__)
	subparsers = parser.add_subparsers()

	for command_name in COMMANDS:
		module = importlib.import_module('.commands.' + command_name, 'sqt')
		subparser = subparsers.add_parser(command_name,
			help=module.__doc__.split('\n')[1], description=module.__doc__)
		subparser.set_defaults(func=module.main)
		module.add_arguments(subparser)

	args = parser.parse_args()
	if not hasattr(args, 'func'):
		parser.error("Please provide a command")
	else:
		args.func(args)


if __name__ == '__main__':
	main()



"""
Quality trimming and filtering.
"""
from cutadapt.qualtrim import quality_trim_index as _qtrimindex

def quality_trim_index(qualities, cutoff, base=33):
	return _qtrimindex(qualities, 0, cutoff, base=base)[1]


def expected_errors(qualities, base=33):
	"""
	Return expected number of errors.

	qualities -- ASCII-encoded qualities (chr(qual + base)).
	"""
	return sum(10 ** (-(ord(q) - base) / 10) for q in qualities)


try:
	from ._helpers import expected_errors
except ImportError:
	pass



import subprocess
from collections import Counter
from io import StringIO

from .io.fasta import FastaReader
from .utils import available_cpu_count

from sqt._helpers import globalalign, hamming_distance


class GlobalAlignment:
	"""A global alignment between two strings"""

	def __init__(self, s, t, semiglobal=False):
		if type(s) is str:
			s = s.encode('ascii')
		if type(t) is str:
			t = t.encode('ascii')
		flags = 15 if semiglobal else 0  # TODO this constant shouldn't be here
		# globalalign uses scores. Set match to 0 to get the same result as with
		# edit distance unless we compute semiglobal alignments, where edit
		# distance does not work.
		match = 1 if semiglobal else 0
		(row1, row2, start1, stop1, start2, stop2, errors) = globalalign(s, t, flags=flags, match=match)
		assert semiglobal or start1 == start2 == 0
		assert semiglobal or stop1 == len(s)
		assert semiglobal or stop2 == len(t)
		assert len(row1) == len(row2)
		self.row1 = row1.decode('ascii')
		self.row2 = row2.decode('ascii')
		self.start1 = start1
		self.stop1 = stop1
		self.start2 = start2
		self.stop2 = stop2
		self.errors = errors
		# Overlap start and stop in alignment coordinates.
		# row1[overlap_start:overlap_stop] is the overlapping alignment.
		self.overlap_start = max(start1, start2)
		self.overlap_stop = len(row1) - max(len(s) - stop1, len(t) - stop2)

	def _replace_gaps(self, row, gap_char):
		start = self.overlap_start
		stop = self.overlap_stop
		front = row[:start].replace('\0', ' ')
		middle = row[start:stop].replace('\0', gap_char)
		end = row[stop:].replace('\0', ' ')
		return front + middle + end

	def print(self, width=100, gap_char='-'):
		"""multi-line output and parameters, therefore not called __str__"""
		row1 = self._replace_gaps(self.row1, gap_char)
		row2 = self._replace_gaps(self.row2, gap_char)
		for i in range(0, len(row1), width):
			top = row1[i:i+width]
			bottom = row2[i:i+width]
			m = []
			for c, d in zip(top, bottom):
				if c == d:
					m.append('|')
				elif c == ' ' or d == ' ':
					m.append(' ')
				else:
					m.append('X')
			middle = ''.join(m)
			if i > 0:
				print()
			print(top)
			print(middle)
			print(bottom)

	def __str__(self):
		return 'GA(row1={}, row2={}, errors={})'.format(
			self.row1, self.row2, self.errors)


def edit_distance(s, t, maxdiff=-1):
	"""
	Return the edit distance between the strings s and t.
	The edit distance is the sum of the numbers of insertions, deletions,
	and mismatches that is minimally necessary to transform one string
	into the other.

	If maxdiff is not -1, then a banded alignment is performed. In that case,
	the true edit distance is returned if and only if it is maxdiff or less.
	Otherwise, a value is returned that is guaranteed to be greater than
	maxdiff, but which is not necessarily the true edit distance.
	"""
	m = len(s) # index: i
	n = len(t) # index: j

	e = maxdiff
	if e != -1 and abs(m - n) > e:
		return abs(m - n)

	# dynamic programming "table" (just a single column)
	# note that using an array('h', ...) here is not faster
	costs = list(range(m+1))

	# calculate alignment (using unit costs)
	for j in range(1, n+1):
		start = 1
		stop = m + 1
		if e != -1: # banded
			stop = min(stop, j + e + 1)
			if j <= e:
				prev = costs[0]
				costs[0] += 1
				start = 1
			else:
				start = j - e
				prev = costs[start-1]
		else:
			prev = costs[0]
			costs[0] += 1
		for i in range(start, stop):
			c = min(
				prev + int(s[i-1] != t[j-1]),
				costs[i] + 1,
				costs[i-1] + 1)
			prev = costs[i]
			costs[i] = c

	return costs[-1]


try:
	from sqt._helpers import edit_distance
except:
	pass


def multialign(sequences, program='mafft', threads=available_cpu_count()):
	"""
	Wrapper for multiple sequence alignment tools. Currently supported are
	* ClustalO (http://www.clustal.org/omega/),
	* MAFFT (http://mafft.cbrc.jp/alignment/software/)
	* MUSCLE (http://www.drive5.com/muscle/)

	A package using the libclustalo library directly also exists:
	https://github.com/benchling/clustalo-python/blob/master/clustalo.c
	It is not for Python 3.

	sequences -- a dictionary mapping names to sequences.
		Use an OrderedDict if order matters.

	program -- must be 'clustalo', 'mafft', 'muscle', 'muscle-medium',
		'muscle-fast'. The latter calls MUSCLE with parameters that make it run
		faster (but less accurate). 'muscle-medium' is in between muscle and
		muscle-fast.

	threads -- number of threads to use for those programs that support it.
		By default, set to the number of processors.
	"""
	if program == 'mafft':
		args = ['mafft', '--quiet', '--thread', str(threads), '-']
	elif program == 'clustalo':
		args = ['clustalo', '--threads='+str(threads), '--infile=-']
	elif program == 'muscle':
		args = ['muscle', '-quiet', '-in', '-', '-out', '-']
	elif program == 'muscle-fast':
		args = ['muscle', '-quiet', '-maxiters', '1', '-diags', '-in', '-', '-out', '-']
	elif program == 'muscle-medium':
		args = ['muscle', '-quiet', '-maxiters', '2', '-diags', '-in', '-', '-out', '-']
	else:
		raise ValueError('program {!r} not supported'.format(program))

	fasta_data = ''.join('>{}\n{}\n'.format(name, seq) for name, seq in sequences.items())
	result = subprocess.check_output(args, input=fasta_data, universal_newlines=True)
	aligned = list(FastaReader(StringIO(result)))

	return { record.name: record.sequence.upper() for record in aligned }


def consensus(aligned, threshold=0.7, ambiguous='N', keep_gaps=False):
	"""
	Compute a consensus from multialign() output.

	Idea taken from BioPython’s SummaryInfo.dumb_consensus function.

	aligned -- a dict mapping names to sequences or a list of sequences
	keep_gaps -- whether the returned sequence contains gaps (-)
	"""
	result = []
	n = len(aligned)
	if hasattr(aligned, 'values'):
		sequences = aligned.values()
	else:
		sequences = aligned
	ambiguous = 'N'
	for chars in zip(*sequences):
		char, freq = Counter(chars).most_common(1)[0]
		if freq / n >= threshold:
			if keep_gaps or char != '-':
				result.append(char)
		else:
			result.append(ambiguous)
	return ''.join(result)



#!/usr/bin/env python3
"""
Print and optionally plot a read length histogram of one or more FASTA or FASTQ
files. If more than one file is given, a total is also printed.
"""
import sys
from collections import Counter
from sqt import SequenceReader, HelpfulArgumentParser

# Make potential import failures happen before we read in files
import matplotlib as mpl
mpl.use('Agg')  # enable matplotlib over an ssh connection without X
import matplotlib.pyplot as plt
import numpy as np
import warnings

__author__ = "Marcel Martin"

warnings.filterwarnings('ignore', 'axes.color_cycle is deprecated and replaced with axes.prop_cycle')
warnings.filterwarnings('ignore', 'The `IPython.html` package')
warnings.filterwarnings('ignore', 'The `IPython.kernel` package')
warnings.filterwarnings('ignore', 'IPython.utils.traitlets has moved')

import seaborn as sns

def add_arguments(parser):
	arg = parser.add_argument
	arg('--zero', default=False, action='store_true', help='Print also rows with a count of zero')

	arg('--plot', default=None, help='Plot to this file (.pdf or .png). '
		'If multiple sequence files given, plot only total.')
	arg('--bins', default=50, type=int, help='Number of bins in the plot. '
		'Default: %(default)s')
	arg('--maxy', default=None, type=float, help='Maximum y in plot')
	arg('--left', default=0, type=float, help='Minimum x in plot')
	arg('--outliers', default=False, action='store_true',
		help='In the plot, summarize outliers greater than the 99.9 percentile '
		'in a red bar.')
	arg('--title', default='Read length histogram of {}',
		help="Plot title. {} is replaced with the input file name. "
		"Default: '%(default)s'")

	arg('seqfiles', nargs='+', metavar='FASTA/FASTQ',
		help='Input FASTA/FASTQ file(s) (may be gzipped).')


def length_histogram(path):
	"""Return a list of lengths """
	lengths = []
	with SequenceReader(path) as reader:
		for record in reader:
			lengths.append(len(record.sequence))
	return lengths


def plot_histogram(lengths, path, title, max_y=None, min_x=0, bins=50, outliers=False):
	"""
	Plot histogram of lengths to path

	If outliers is True, then the lengths greater than the 99.9 percentile are
	marked separately with bar colored in red.
	"""
	lengths = np.array(lengths)
	if outliers:
		histomax = int(np.percentile(lengths, 99.9) * 1.01)
	else:
		histomax = int(max(lengths, default=100))
	larger = sum(lengths > histomax)

	fig = plt.figure(figsize=(20/2.54, 10/2.54))
	ax = fig.gca()
	ax.set_xlabel('Read length')
	ax.set_ylabel('Frequency')
	ax.set_title(title)
	_, borders, _ = ax.hist(lengths, bins=bins, range=(min_x, histomax))
	if outliers:
		w = borders[1] - borders[0]
		ax.bar([histomax], [larger], width=w, color='red')
		ax.set_xlim(min_x, histomax + 1.5 * w)
	if max_y is not None:
		ax.set_ylim(0, max_y)
	fig.set_tight_layout(True)
	fig.savefig(path)


def main(args=None):
	if args is None:
		parser = HelpfulArgumentParser(description=__doc__)
		add_arguments(parser)
		args = parser.parse_args()
	overall_lengths = []
	for path in args.seqfiles:
		print("## File:", path)
		print("length", "frequency", sep='\t')
		lengths = length_histogram(path)
		freqs = Counter(lengths)
		for length in range(0, max(freqs, default=100) + 1):
			freq = freqs[length]
			if args.zero or freq > 0:
				print(length, freq, sep='\t')
		overall_lengths.extend(lengths)

	if len(args.seqfiles) > 1:
		print("## Total")
		print("length", "frequency", sep='\t')
		freqs = Counter(overall_lengths)
		for length in range(0, max(freqs) + 1):
			freq = freqs[length]
			if args.zero or freq > 0:
				print(length, freq, sep='\t')
		title = args.title.format('{} input files'.format(len(args.seqfiles)))
	else:
		title = args.title.format(args.seqfiles[0])
	if args.plot:
		plot_histogram(overall_lengths, args.plot, title, args.maxy, min_x=args.left, bins=args.bins, outliers=args.outliers)


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
"""
Compare two strings
"""
from sqt import HelpfulArgumentParser
import sys
import os
from collections import namedtuple
from cutadapt import align
from sqt import FastaReader
from sqt.align import globalalign, GlobalAlignment
from sqt.dna import reverse_complement

__author__ = "Marcel Martin"


Alignment = namedtuple('Alignment', 'start1 stop1 start2 stop2 matches errors')

def add_arguments(parser):
	arg = parser.add_argument
	arg('--semiglobal', '--overlap', action='store_true', default=False,
		help='Run a semi-global alignment (for detecting overlaps). '
		'Default: global alignment.')
	arg('--max-error-rate', '-e', type=float, default=None,
		help='Switch to cutadapt algorithm (also enables --semiglobal). '
		'No alignment will be printed.')
	arg('--reverse-complement', '--rc', action='store_true',
		default=False,
		help='Run the alignment also with the reverse-complement of the second '
		'sequence')
	arg('--merge', action='store_true', default=False,
		help='Output a merged sequence (also enables --semiglobal)')
	arg('sequence1', help='Sequence or path to FASTA file. If FASTA, only the first sequence is used.')
	arg('sequence2', help='Sequence or path to FASTA file. If FASTA, only the first sequence is used.')


def print_numbers(sequence1, sequence2, alignment, overlap):
	print()
	print('Length of sequence 1:', len(sequence1))
	print('Length of sequence 2:', len(sequence2))
	print('Errors in alignment:', alignment.errors)
	if overlap:
		print('Length of overlap in sequence 1:', alignment.stop1 - alignment.start1)
		print('Length of overlap in sequence 2:', alignment.stop2 - alignment.start2)
	if hasattr(alignment, 'matches'):
		print('Matches:', alignment.matches)


def load_sequence(path_or_sequence):
	if os.path.exists(path_or_sequence):
		with FastaReader(path_or_sequence) as fr:
			sequence = next(iter(fr)).sequence
	else:
		sequence = path_or_sequence
	return sequence


def main(args=None):
	if args is None:
		parser = HelpfulArgumentParser(description=__doc__)
		add_arguments(parser)
		args = parser.parse_args()
	sequence1 = load_sequence(args.sequence1)
	sequence2 = load_sequence(args.sequence2)

	sequences = [ (False, sequence2) ]
	if args.reverse_complement:
		sequences.append((True, reverse_complement(sequence2)))

	if args.merge:
		args.semiglobal = True

	# credit: http://stackoverflow.com/questions/566746/
	rows, columns = os.popen('stty size', 'r').read().split()
	columns = int(columns)

	for revcomp, sequence2 in sequences:
		if revcomp:
			print('Alignment with reverse-complemented sequence 2:')
		else:
			print('Alignment:')
		print()
		if args.max_error_rate is not None:
			flags = align.SEMIGLOBAL
			result = align.locate(sequence1.upper(), sequence2.upper(), max_error_rate=args.max_error_rate, flags=flags)
			if result is None:
				print('No alignment found')
				continue
			alignment = Alignment(*result)
			print_numbers(sequence1, sequence2, alignment, overlap=True)
		else:
			alignment = GlobalAlignment(sequence1.upper(), sequence2.upper(), semiglobal=args.semiglobal)
			alignment.print(width=columns, gap_char='-')
			print_numbers(sequence1, sequence2, alignment, args.semiglobal)

		if args.merge:
			merged = sequence1[0:alignment.start1]
			merged += sequence2[0:alignment.start2]
			merged += sequence1[alignment.start1:alignment.stop1]
			merged += sequence1[alignment.stop1:]
			merged += sequence2[alignment.stop2:]
			print('Merged (length {}):'.format(len(merged)), merged)


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
"""
Read in a file with variants (typically VCF) and check if the given
reference allele matches what is actually in the reference sequence
at the given position.
"""
import sys
import csv

from sqt import HelpfulArgumentParser, IndexedFasta

__author__ = "Marcel Martin"


def main():
	parser = HelpfulArgumentParser(description=__doc__)
	arg = parser.add_argument
	arg("reference", help="Reference file in FASTA format. An associated .fai index file must exist.")
	arg("variantfile", help="Input (VCF) file with variants.")
	args = parser.parse_args()

	reference = IndexedFasta(args.reference)
	with open(args.variantfile) as f:
		reader = csv.reader([row for row in f if row[0] != '#'], delimiter='\t')
		n = 0
		for row in reader:
			n += 1
			chrom = row[0]
			if chrom.startswith('chr'):
				chrom = chrom[3:]
			pos = int(row[1]) - 1
			ref = row[3]
			actual = reference.get(chrom)[pos:pos+len(ref)].decode()
			if not actual == ref:
				print('Problem with record no.', n, 'found:')
				print('CHROM={} POS={} REF={}'.format(chrom, pos+1, ref))
				print('Reference sequence in FASTA file is:', actual)
				sys.exit(1)
		else:
			print((n, 'records checked, everything ok.'))
			sys.exit(0)


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
"""
Extract all reads from a BAM file that map to a certain location, but try hard
to extract them even if hard clipping is used.

TODO reverse-complementarity is ignored
TODO behavior when no region is given is not well documented
"""
__author__ = 'Marcel Martin'

import sys
import logging
from pysam import Samfile
from sqt import HelpfulArgumentParser, FastqWriter
from sqt.region import Region
from sqt.reads import AlignedRead
from sqt.cigar import Cigar

logger = logging.getLogger(__name__)

def add_arguments(parser):
	arg = parser.add_argument
	arg("--missing-quality", type=int, default=40,
		help='Quality value to use if an entry does not have qualities '
			'(default: %(default)s)')
	arg("bam", metavar="SAM/BAM", help="Name of a SAM or BAM file")
	arg("-L", '--bed', metavar="BED-FILE", action='append',
		help="BED file with regions", default=[])
	arg("region", nargs='*', help="Region")


def extract_read(aligned_read, aligned_segment, samfile):
	"""
	Return the primary AlignedSegment (the one that is not hard clipped) for
	the aligned read.
	"""
	def is_hard_clipped(segment):
		cig = Cigar(segment.cigar)
		return cig.hard_clipping_left != 0 or cig.hard_clipping_right != 0

	if not is_hard_clipped(aligned_segment):
		return aligned_segment
	for supplementary in aligned_read:
		refname, pos = supplementary.reference_name, supplementary.pos
		for segment in samfile.fetch(refname, pos, pos+1, multiple_iterators=True):
			if (segment.query_name == aligned_segment.query_name and
					not is_hard_clipped(segment)):
				return segment
	return None


def parse_bed(file):
	"""
	Yield Region objects for each line of the BED-formatted input file.
	If a line in the BED file contains less than three fields, the end coordinate
	is set to None. If it contains only one field, the start coordinate is set
	to zero.
	"""
	for line in file:
		line = line.strip()
		if line.startswith('#'):
			continue
		fields = line.split('\t')
		if len(fields) == 1:
			start = 0
			stop = None
		else:
			start = int(fields[1])
			if len(fields) == 2 or fields[2] == '':
				stop = None
			else:
				stop = int(fields[2])
		yield Region(fields[0], start, stop)



def collect_regions(region_specifications, bedpaths):
	"""
	Special case: if neither specifications nor bedpaths are given, yield a
	single None.
	"""
	if not region_specifications and not bedpaths:
		yield None
		return
	for path in bedpaths:
		with open(path) as f:
			for region in parse_bed(f):
				yield region
	for spec in region_specifications:
		yield Region(spec)


def main(args=None):
	if args is None:
		parser = HelpfulArgumentParser(description=__doc__)
		add_arguments(parser)
		args = parser.parse_args()
	missing_quality = chr(args.missing_quality + 33)
	written_reads = set()
	not_found = set()
	no_qualities = 0
	indirect = 0
	n_regions = 0
	names = set()
	regions = list()
	with FastqWriter(sys.stdout) as writer, Samfile(args.bam) as sf:
		for region in collect_regions(args.region, args.bed):
			if region is None:
				region_iter = sf
			else:
				region_iter = sf.fetch(region.reference, region.start, region.stop)
			n_regions += 1
			for record in region_iter:
				names.add(record.query_name)
				if region is None:
					if record.is_supplementary or record.is_secondary or record.is_unmapped:
						continue
					if record.mapping_quality < 1:
						continue
					if record.query_alignment_length < 1000:
						continue
					segment = record
				else:
					if record.is_unmapped:
						assert False, 'shouldn’t happen'
						continue
					assert record.cigar is not None
					if record.query_name in written_reads or record.query_name in not_found:
						continue
					aligned_read = AlignedRead(record, sf.getrname(record.tid))
					segment = extract_read(aligned_read, record, sf)
					if segment is None:
						not_found.add(record.query_name)
						continue
					if segment is not record:
						indirect += 1
					assert segment.query_name == record.query_name
				if segment.query_qualities is not None:
					qualities = ''.join(chr(c+33) for c in segment.query_qualities)
				else:
					qualities = missing_quality * len(segment.query_sequence)
					no_qualities += 1
				writer.write(segment.query_name, segment.query_sequence, qualities)
				written_reads.add(record.query_name)

	logger.info('%s unique read names in %s region(s)', len(names), n_regions)
	logger.info('%s entries written (%s found indirectly)', len(written_reads), indirect)
	logger.info('%s without qualities', no_qualities)


if __name__ == '__main__':
	main()



"""
Reading 454 .sff files (flowgram file format)
(c) Sven Rahmann, 2011
see: http://www.ncbi.nlm.nih.gov/Traces/trace.cgi?cmd=show&f=formats&m=doc&s=format
"""
__author__ = 'Sven Rahmann'

from struct import calcsize, unpack
from collections import namedtuple

class FormatError(Exception):
	pass


class SFFFile():
	_MAGIC = 0x2E736666

	def __init__(self, filename):
		self.filename = filename
		with open(filename, mode="rb") as f:
			self._set_info_from_header(f)

	def _process_padding(self, f, p):
		if not (0 <= p < 8):
			raise FormatError("Padding mismatch, padding=" + str(p))
		padder = f.read(p)
		if padder.count(b'\0') != p:
			raise FormatError("Padding seems to contain data")

	def _fread(self, f, fmt):
		b = calcsize(fmt)
		data = f.read(b)
		if len(data) < b:
			raise FormatError("chunk for " + fmt + " too short: " + str(len(data)) + "/" + str(b))
		return unpack(fmt, data)


	def _set_info_from_header(self, f):
		"""read the sff header and store its information in self"""
		_FIXEDLEN = 31
		# Read file header (constant part), length 31 bytes, 9 fields
		# big endian encdoing        >
		# magic_number               I  == _MAGIC
		# version                    4B == (0,0,0,1)
		# index_offset               Q
		# index_length               I
		# number_of_reads            I
		# header_length              H  divisble by 8
		# key_length                 H
		# number_of_flows_per_read   H
		# flowgram_format_code       B == 1
		headerformat = '>I4BQIIHHHB'
		assert calcsize(headerformat) == _FIXEDLEN
		(magic_number,
			ver0, ver1, ver2, ver3,
			index_offset, index_length,
			number_of_reads,
			header_length,
			key_length,
			number_of_flows_per_read,
			flowgram_format_code) = self._fread(f, headerformat)
		if magic_number != SFFFile._MAGIC:
			raise FormatError("Magic number is {} instead of {}").format(magic_number, SFFFile._MAGIC)
		if (ver0, ver1, ver2, ver3) != (0, 0, 0, 1):
			raise FormatError("Unsupported .sff version ({}.{}.{}.{})".format(ver0, ver1, ver2, ver3))
		if (index_offset != 0) ^ (index_length != 0):
			raise FormatError("Index offset is {}, but length is {}".format(index_offset, index_length))
		if (index_offset % 8 != 0) or (index_length % 8 != 0):
			#raise FormatError("Index (offset, length) must be divisible by 8"+str(index_offset)+","+str(index_length))
			pass
		if header_length % 8 != 0:
			raise FormatError("Header length must be divisible by 8, but is {}".format(header_length))
		if flowgram_format_code != 1:
			raise FormatError("Flowgram format code {} not supported".format(flowgram_format_code))
		# Read variable part of header:
		# flow_chars     {number_of_flows_per_read}s
		# key_sequence   {key_length}s
		flow_chars = f.read(number_of_flows_per_read)
		key_sequence = f.read(key_length)
		# padding        *B
		padding = header_length - number_of_flows_per_read - key_length - _FIXEDLEN
		self._process_padding(f, padding)
		# set attributes:
		self.magic_number = magic_number
		self.version = (ver0, ver1, ver2, ver3)
		self.has_index = (index_offset != 0) and (index_length != 0)
		self.index_offset = index_offset
		self.index_length = index_length
		self.number_of_reads = number_of_reads
		self.header_length = header_length
		#self.key_length   = key_length  #  == len(key_sequence)
		self.key_sequence = key_sequence
		self.number_of_flows_per_read = number_of_flows_per_read
		self.flow_chars = flow_chars
		self.flowgram_format_code = flowgram_format_code

	# generator function to iterate over all reads
	def reads(self):
		"""yields each read in this .sff file as a Read object."""
		header_length = self.header_length
		with open(self.filename, mode="rb") as f:
			f.read(header_length)
			fpos = header_length
			checkindex = self.has_index
			for r in range(self.number_of_reads):
				assert fpos % 8 == 0, "file position {} not divisible by 8".format(fpos)
				if checkindex and fpos == self.index_offset:
					f.read(index_length)
					fpos += index.length
					checkindex = False
				read = self._next_read(f, r)
				yield read

	def __iter__(self):
		return self.reads()

	def _next_read(self, f, readindex):
		"""reads the next read record from an open .sff file f"""
		# Read fixed part of read header, 7 fields
		# read_header_length     H
		# name_length            H
		# seq_len                I
		# clip_qual_left         H
		# clip_qual_right        H
		# clip_adapter_left      H
		# clip_adapter_right     H
		header_fmt = ">HHIHHHH"
		(read_header_length,
		name_length,
		seq_len,
		clip_qual_left, clip_qual_right,
		clip_adapter_left, clip_adapter_right) = self._fread(f, header_fmt)
		# check format
		expected = ((16 + name_length + 7) // 8) * 8
		if read_header_length != expected:
			raise FormatError("read header length should be 16 + name length, rounded up mod 8")
		# name                 c * name_length
		# padding              B * [to fill]
		name_fmt = ">" + str(name_length) + "s"
		(namebytes,) = self._fread(f, name_fmt)
		name = namebytes.decode()
		padding = read_header_length - (16 + name_length)
		self._process_padding(f, padding)
		# flowgram_values      H * nflows  [type may differ with future formats]
		# flow_index_per_base  B * seq_len
		# bases                c * seq_len
		# quality_scores       B * seq_len
		# padding              B * [to fill]
		datatypes = ".H"
		flow_fmt = ">" + str(self.number_of_flows_per_read) + datatypes[self.flowgram_format_code]
		byte_fmt = ">" + str(seq_len) + "B"
		char_fmt = ">" + str(seq_len) + "s"
		flowgram_values = self._fread(f, flow_fmt)
		flow_index_per_base = self._fread(f, byte_fmt)
		if seq_len > 0:
			flow_index_per_base = (flow_index_per_base[0] - 1,) + flow_index_per_base[1:] # convert to 0-based index
		(bases,) = self._fread(f, char_fmt)
		quality_scores = self._fread(f, byte_fmt)
		datalen = sum(map(calcsize, (flow_fmt, byte_fmt, char_fmt, byte_fmt)))
		padding = (-datalen) % 8
		self._process_padding(f, padding)
		r = Read(readindex, name, self, # self is the current sff file
				clip_qual_left, clip_qual_right, clip_adapter_left, clip_adapter_right,
				flowgram_values, bases, flow_index_per_base, quality_scores)
		return r


# TODO move the Flow and Flowgram class into some other module?
class Flow(namedtuple('Flow', ['char', 'intensity'])):
	def __str__(self):
		return '{}{:.2f}'.format(chr(self.char), self.intensity)


class Flowgram:
	"""
	A flowgram is mathematically defined to be a list of (char, intensity) pairs,
	but this implementation stores two lists, one with chars and one
	with intensities. Indexed access and iteration are supported, however, and result in
	(char, intensity) pairs.
	"""
	def __init__(self, flowchars, intensities):
		if len(flowchars) != len(intensities):
			raise ValueError("lengths of flowchars and intensities must be equal")
		self.flowchars = flowchars
		self.intensities = intensities

	def __iter__(self):
		return (Flow(*flow) for flow in zip(self.flowchars, self.intensities))

	def __len__(self):
		return len(self.flowchars)

	def __str__(self):
		return ' '.join(str(flow) for flow in self)


class Read(Flowgram):
	"""
	Represent a flowgram read by the following attributes:
	- index: int, 0-based running number of the read in the sff file
	- name: bytes, a unique id for the read
	- sff: SFFFile, from which the read was obtained
	- flowvalues: tuple of ints
	- flowchars:  bytes, usually  b'TACG' * n for some n
	- qual: tuple of ints, quality values (10log10-representation)
	- clip: 4-tuple with clipping information
	- key: bytes, the key sequence (usually b'TCAG')
	- bases: bytes, the 454-converted sequence representation of the flowvalues
	- flow_index_per_base: tuple of ints
	"""
	def __init__(self, index, name, sff_origin,
				clip_qual_left, clip_qual_right, clip_adapter_left, clip_adapter_right,
				flowgram_values, bases, flow_index_per_base, quality_scores):
		self.index = index
		self.name = name
		self.sff = sff_origin
		self.clip = dict(ql=clip_qual_left, qr=clip_qual_right,
						al=clip_adapter_left, ar=clip_adapter_right)
		self.flowchars = sff_origin.flow_chars
		self.flowvalues = flowgram_values
		self.intensities = [ v / 100.0 for v in flowgram_values ]
		self.key = sff_origin.key_sequence
		self.bases = bases
		self.flow_index_per_base = flow_index_per_base
		self.qual = quality_scores

		self.insert_start = max(1, max(self.clip['ql'], self.clip['al'])) - 1
		self.insert_stop = min(len(bases) if self.clip['qr'] == 0 else self.clip['qr'],
			len(bases) if self.clip['ar'] == 0 else self.clip['ar'])


class ClippedRead(Flowgram):
	"""
	A read in which both the bases and the flowgram (flowchars and intensities) are
	clipped to the insert region.

	A ClippedRead also has a base_to_flow attribute which maps base
	indices to flow indices.
	"""

	def __init__(self, read, start=0, stop=None):
		"""
		read is a Read object.
		start and stop specify additional offsets within the insert region.
		start and stop are interpreted as in slice notation (negative values
		and None are allowed).
		"""
		self._read = read
		self.name = read.name
		# _fl prefix: index into flowgram
		# _b  prefix: index into bases
		insert_start_b = max(1, max(read.clip['ql'], read.clip['al'])) - 1
		insert_stop_b = min(len(read.bases) if read.clip['qr'] == 0 else read.clip['qr'],
			len(read.bases) if read.clip['ar'] == 0 else read.clip['ar'])
		start_b, stop_b, _ = slice(start, stop).indices(insert_stop_b - insert_start_b)
		start_b += insert_start_b
		stop_b += insert_start_b
		b2fl = self._compute_base_to_flow()
		start_fl = b2fl[start_b]
		stop_fl = b2fl[stop_b]

		self.bases = read.bases[start_b:stop_b]
		self.intensities = read.intensities[start_fl:stop_fl]
		self.flowchars = read.flowchars[start_fl:stop_fl]
		self.base_to_flow = [ index_fl - start_fl for index_fl in b2fl[start_b:stop_b] ]
		assert len(self.bases) == len(self.base_to_flow)

	def _compute_base_to_flow(self):
		"""
		map a base index to a flow index
		"""
		m = []
		v = 0
		for index in self._read.flow_index_per_base:
			v += index
			m.append(v)
		assert v < len(self._read.intensities)
		m.append(len(self._read.intensities))
		return m



"""
Model an interval on a reference.
"""
class Region:
	def __init__(self, specification, start=None, stop=None, reverse_complement=False):
		"""
		specification -- description of the region as a string, such as
		"chr14:22-111"

		If start is given, the specification is considered to be a reference and
		the parameters start, stop, reverse_complement are used directly.
		"""
		if start is None:
			self.reference, self.start, self.stop, self.is_reverse_complement = self._parse_region(specification)
		else:
			self.reference = specification
			self.start = start
			self.stop = stop
			self.is_reverse_complement = reverse_complement

	@staticmethod
	def _parse_region(s):
		"""
		Parse a string like "name:begin-end" or "name:begin..end".
		The returned tuple is (name, start, stop, revcomp).
		start is begin-1, stop is equal to end.
		That is, this function converts from 1-based intervals to pythonic
		open intervals!

		The string may be prefixed with "rc:", in which case revcomp is set to True.

		If 'end' is an empty string (as in "chrx:1-"), then stop is set to None.
		If no range is given, as in "chrx:27", then stop is set to start+1.
		If only 'name' is given (or "rc:name"), start is set to 0 and stop to None.

		Commas within the numbers (thousands separators) are ignored.
		"""
		revcomp = False
		if s.startswith('rc:'):
			revcomp = True
			s = s[3:]
		fields = s.rsplit(':', 1)
		if len(fields) == 1:
			region = (fields[0], 0, None, revcomp)
		else:
			if '..' in fields[1]:
				sep = '..'
			else:
				sep = '-'
			coords = fields[1].split(sep, maxsplit=1)
			start = int(coords[0].replace(',', ''))
			if len(coords) == 1:
				stop = start
			else:
				stop = int(coords[1].replace(',', '')) if coords[1] != '' else None
			assert 0 < start and (stop is None or start <= stop)
			region = (fields[0], start-1, stop, revcomp)
		return region

	def __str__(self):
		"""

		"""
		prefix = 'rc:' if self.is_reverse_complement else ''
		if self.start == 0 and self.stop is None:
			return prefix + self.reference
		if self.start + 1 == self.stop:
			return "{}{}:{}".format(prefix, self.reference, self.start+1)
		stop = '' if self.stop is None else self.stop
		return "{}{}:{}-{}".format(prefix, self.reference, self.start+1, stop)

	def __repr__(self):
		return "Region({!r})".format(str(self))

	def __eq__(self, other):
		return (self.reference == other.reference and self.start == other.start
		  and self.stop == other.stop
		  and self.is_reverse_complement == other.is_reverse_complement)



#!/usr/bin/env python3
"""
- Python 2 and 3 compatible fast reverse complement
- definition of genetic code (triplets to amino acids)
- fast translation of nucleotide strings to amino acids
- other misc. functions
"""
import sys
import re
import random
from ._codons import GENETIC_CODE  # for re-export
from sqt._helpers import nt_to_aa


if sys.version < '3':
	from string import maketrans
else:
	maketrans = bytes.maketrans
	_TR_STR = str.maketrans('ACGTUMRWSYKVHDBNacgtumrwsykvhdbn', 'TGCAAKYWSRMBDHVNtgcaakywsrmbdhvn')

_TR = maketrans(b'ACGTUMRWSYKVHDBNacgtumrwsykvhdbn', b'TGCAAKYWSRMBDHVNtgcaakywsrmbdhvn')

if sys.version < '3':
	def reverse_complement(s):
		return s.translate(_TR)[::-1]
else:
	def reverse_complement(s):
		if isinstance(s, str):
			return s.translate(_TR_STR)[::-1]
		else:
			return s.translate(_TR)[::-1]


AMINO_ACID_REGEXES = dict(
	A='GC[ACGT]',
	C='TG[CT]',
	D='GA[CT]',
	E='GA[AG]',
	F='TT[CT]',
	G='GG[ACGT]',
	H='CA[CT]',
	I='AT[ACT]',
	K='AA[AG]',
	L='(CT[ACGT]|TT[AG])',
	M='ATG',
	N='AA[CT]',
	P='CC[ACGT]',
	Q='CA[AG]',
	R='(AG[AG]|CG[ACGT])',
	S='AG[CT]|TC[ACGT]',
	T='AC[ACGT]',
	V='GT[ACGT]',
	W='TGG',
	Y='TA[CT]',
	X='[ACGT][ACGT][ACGT]'
)


def amino_acid_regex(aa_sequence, compile=False):
	"""
	Given an amino acid sequence, return a regular expression that can be used
	to match a nucleotide sequence. If compile is True, the regular expression
	is compiled with re.compile, otherwise the regex is returned as a string.
	"""
	regex = ''.join(AMINO_ACID_REGEXES[aa] for aa in aa_sequence)
	return re.compile(regex) if compile else regex


def n_intervals(sequence, N='N'):
	"""
	Given a sequence, yield all intervals containing only N characters as
	tuples (start, stop). If the sequence is a bytes/bytearray object,
	set N=ord(b'N')

	>>> list(n_intervals('ACGTnNAC'))
	[(4, 6)]
	>>> list(n_intervals(b'ACGTnNAC', N=ord(b'N')))
	[(4, 6)]
	"""
	sequence = sequence.upper()
	start = sequence.find(N)
	while start >= 0:
		stop = start + 1
		while stop < len(sequence) and sequence[stop] == N:
			stop += 1
		yield (start, stop)
		start = sequence.find(N, stop)


def intervals_complement(intervals, length):
	"""
	Given an iterable of sorted, nonoverlapping intervals as (start, stop)
	pairs, yield the complementary intervals. The result is equivalent to
	[(0, length)] minus the given intervals.

	>>> list(intervals_complement([(1, 2), (4, 6)], length=10))
	[(0, 1), (2, 4), (6, 10)]
	"""
	prev_stop = 0
	for start, stop in intervals:
		if start >= length:
			break
		if prev_stop != start:
			yield (prev_stop, start)
		prev_stop = stop
	if prev_stop < length:
		yield (prev_stop, length)


def mutate(seq, substitution_rate=0.1, indel_rate=0.01, alphabet='ACGT', counts=False):
	"""
	Mutate a DNA sequence by randomly introducing substitutions and indels.
	This does not use a very sophisticated model.

	Return the mutated sequence.

	substitution_rate -- The probability at which an individual base is substituted with a different one.

	indel_rate -- At any position, a random base is inserted at probability indel_rate/2, and
		the base is deleted at probability indel_rate/2.


	counts -- If this is True, a triple (mutated_sequence, number_of_substitutions, number_of_indels)
		is returned instead of just the sequence.
	"""
	other_chars = {}
	for c in alphabet:
		other_chars[c] = alphabet.replace(c, '')
	mutated = []
	n_sub = 0
	n_indel = 0
	indels = indel_rate > 0
	for c in seq:
		c = c.upper()
		r = random.random()
		if r < substitution_rate:
			# mutate base
			d = random.choice(other_chars.get(c, alphabet))
			mutated.append(d)
			n_sub += 1
		elif indels and r < (substitution_rate + 0.5 * indel_rate):
			# insertion
			mutated.append(random.choice(alphabet))
			mutated.append(c)
			n_indel += 1
		elif indels and r < (substitution_rate + indel_rate):
			n_indel += 1
			# deletion
			pass
		else:
			# no change
			mutated.append(c)
	if counts:
		return ''.join(mutated), n_sub, n_indel
	else:
		return ''.join(mutated)



#!/usr/bin/env python3
"""
Mutate nucleotides in a FASTA or FASTQ file

The modified sequences are written to standard output.
"""
import sys
import random
from sqt import HelpfulArgumentParser
from sqt.io.fasta import FastaReader, FastaWriter
from sqt.dna import mutate

__author__ = "Marcel Martin"


def main():
	parser = HelpfulArgumentParser(description=__doc__)
	parser.add_argument("--rate", type=float, default=0.03,
		help="Substitution rate (default: %(default)s)")
	parser.add_argument("--indel-rate", type=float, default=0.0005,
		help="Indel rate (default: %(default)s)")
	parser.add_argument("--seed", type=int, default=None,
		help="Set random seed for reproducible runs (default: use different seed each run)")
	parser.add_argument("fasta", metavar='FASTA-file', help="Input FASTA file")
	args = parser.parse_args()

	if args.seed is not None:
		random.seed(args.seed)
	fasta_output = FastaWriter(sys.stdout, line_length=80)
	for record in FastaReader(args.fasta):
		mutated, n_sub, n_indel = mutate(record.sequence, rate=args.rate, alphabet='ACGT', indel_rate=args.indel_rate, counts=True)
		fasta_output.write(record.name + '-sub{}-indel{}'.format(n_sub, n_indel), mutated)


if __name__ == '__main__':
	main()



# kate: syntax Python;
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from collections import Counter
from cython.view cimport array as cvarray
import cython

from ._codons import GENETIC_CODE


DEF START_WITHIN_SEQ1 = 1
DEF START_WITHIN_SEQ2 = 2
DEF STOP_WITHIN_SEQ1 = 4
DEF STOP_WITHIN_SEQ2 = 8
DEF SEMIGLOBAL = 15
DEF ALLOW_WILDCARD_SEQ1 = 1
DEF ALLOW_WILDCARD_SEQ2 = 2

DEF INSERTION_COST = 1
DEF DELETION_COST = 1
DEF MATCH_COST = 0
DEF MISMATCH_COST = 1
DEF WILDCARD_CHAR = b'N'

# structure for a DP matrix entry
ctypedef struct ScoreEntry:
	int score
	int backtrace

# insertion means: inserted into seq1 (does not appear in seq2)

DEF GAPCHAR = b'\0'


@cython.boundscheck(False)
def edit_distance(s, t, int maxdiff=-1):
	"""
	Return the edit distance between the strings s and t.
	The edit distance is the sum of the numbers of insertions, deletions,
	and mismatches that is minimally necessary to transform one string
	into the other.

	If maxdiff is not -1, then a banded alignment is performed. In that case,
	the true edit distance is returned if and only if it is maxdiff or less.
	Otherwise, a value is returned that is guaranteed to be greater than
	maxdiff, but which is not necessarily the true edit distance.
	"""
	cdef int m = len(s)  # index: i
	cdef int n = len(t)  # index: j
	cdef int e = maxdiff
	cdef int i, j, start, stop, c, prev, smallest
	cdef bint match
	cdef bytes s_bytes, t_bytes
	cdef char* sv
	cdef char* tv


	# Return early if string lengths are too different
	if e != -1 and abs(m - n) > e:
		return abs(m - n)

	s_bytes = s.encode() if isinstance(s, unicode) else s
	t_bytes = t.encode() if isinstance(t, unicode) else t
	sv = s_bytes
	tv = t_bytes

	# Skip identical prefixes
	while m > 0 and n > 0 and sv[0] == tv[0]:
		sv += 1
		tv += 1
		m -= 1
		n -= 1

	# Skip identical suffixes
	while m > 0 and n > 0 and sv[m-1] == tv[n-1]:
		m -= 1
		n -= 1

	cdef int[:] costs = cvarray(shape=(m+1,), itemsize=sizeof(int), format="i")
	if e == -1:
		# Regular (unbanded) global alignment
		with nogil:
			for i in range(m + 1):
				costs[i] = i

			# compute columns of the alignment matrix (using unit costs)
			prev = 0
			for j in range(1, n+1):
				prev = costs[0]
				costs[0] += 1
				for i in range(1, m+1):
					match = sv[i-1] == tv[j-1]
					c = min(
						prev + 1 - match,
						costs[i] + 1,
						costs[i-1] + 1)
					prev = costs[i]
					costs[i] = c
	else:
		# Banded alignment
		with nogil:
			for i in range(m + 1):
				costs[i] = i
			smallest = 0
			for j in range(1, n + 1):
				stop = min(j + e + 1, m + 1)
				if j <= e:
					prev = costs[0]
					costs[0] += 1
					smallest = costs[0]
					start = 1
				else:
					start = j - e
					prev = costs[start - 1]
					smallest = maxdiff + 1
				for i in range(start, stop):
					match = sv[i-1] == tv[j-1]
					c = min(
						prev + 1 - match,
						costs[i] + 1,
						costs[i-1] + 1)
					prev = costs[i]
					costs[i] = c
					smallest = min(smallest, c)
				if smallest > maxdiff:
					break
		if smallest > maxdiff:
			return smallest
	return costs[m]


#@cython.boundscheck(False)
def globalalign(char* s1, char* s2, int flags=0, int match=1, int mismatch=-2, int insertion=-2, int deletion=-2):
	"""
	Compute an optimal global or semiglobal alignment between strings s1 and s2.
	An alignment is optimal if it has maximal score.

	The optimal score is not returned. Instead, the number of errors is computed
	and returned.

	Return ... -> (r1, r2, start1, stop1, start2, stop2, errors)

	TODO

	This is a direct translation of the C code and should be re-written to make
	it more readable. (Use Cython's memoryview for the matrix, avoid pointer-like
	access to p1 and p2.)

	FIXME THE REMAINDER OF THIS DOCSTRING

	Return a tuple (row1, row2, start1, stop1, start2, stop2, errors)
	where row1 and row2 are strings of the same length containing the alignment
	(an INDEL is marked by a null byte ('\\0').

	start1 is the position within row1 at which the part of s1, that is aligned, starts.
	start2 is the position within row1 at which the part of s1, that is aligned, ends.
	The same holds for start2, stop2.

	It is always the case that at least one of start1 and start2 is zero.

	It is always the case that either stop1==len(row1) or stop2==len(row2) or both
	(note that len(row1)==len(row2)). This is a property of semiglobal alignments.

	errors is the number of errors in the alignment.

	For example, globalalign("SISSI", "MISSISSIPPI") returns:

	row1 = [   0,   0,   0, 'S', 'I', 'S', 'S', 'I',   0,   0,   0]
	row2 = [ 'M', 'I', 'S', 'S', 'I', 'S', 'S', 'I', 'P', 'P', 'I']
	start1, stop1 = 0, 5
	start2, stop2 = 3, 8
	errors = 0

	This corresponds to the following alignment:
	   SISSI
	   |||||
	MISSISSIPPI
	"""
	cdef int m = len(s1)
	cdef int n = len(s2)
	# DP Matrix:
	#            s2 (j)
	#         ----------> n
	#        |
	# s1 (i) |
	#        |
	#        V
	#        m

	# direction constants for backtrace table
	cdef int LEFT = 1, UP = 2, DIAG = 3

	# the DP matrix is stored column-major
	cdef ScoreEntry[:,:] columns = cvarray(shape=(m+1, n+1), itemsize=sizeof(ScoreEntry), format="ii")

	cdef int i, j, bt, score, tmp

	# initialize first column
	for i in range(m + 1):
		columns[i, 0].score = 0 if (flags & START_WITHIN_SEQ1) else i * deletion
		columns[i, 0].backtrace = UP

	# initialize first row
	for j in range(n + 1):
		columns[0, j].score = 0 if (flags & START_WITHIN_SEQ2) else j * insertion
		columns[0, j].backtrace = LEFT

	# fill the entire DP matrix
	# outer loop goes over columns
	for j in range(1, n+1):
		for i in range(1, m+1):
			bt = DIAG
			score = columns[i-1,j-1].score + (match if (s1[i-1] == s2[j-1]) else mismatch)
			tmp = columns[i-1,j].score + insertion
			if tmp > score:
				bt = UP
				score = tmp
			tmp = columns[i,j-1].score + deletion
			if tmp > score:
				bt = LEFT
				score = tmp
			columns[i,j].score = score
			columns[i,j].backtrace = bt

	# initialize best score and its position to the bottomright cell
	cdef int best_i = m  # also: s1stop
	cdef int best_j = n  # also: s2stop
	cdef int best = columns[m,n].score

	if flags & STOP_WITHIN_SEQ2:
		# search also in last row
		for j in range(n + 1):
			if columns[m,j].score >= best:
				best = columns[m,j].score
				best_i = m
				best_j = j

	cdef ScoreEntry* last_column
	if flags & STOP_WITHIN_SEQ1:
		# search also in last column
		#last_column = &(columns[0,n])
		for i in range(m + 1):
			if columns[i,n].score >= best:
				best_i = i
				best_j = n
				best = columns[i,n].score

	# trace back
	cdef char* alignment1 = <char*>PyMem_Malloc((m+n+4)*sizeof(char))
	if not alignment1:
		raise MemoryError()
	cdef char* alignment2 = <char*>PyMem_Malloc((m+n+4)*sizeof(char))
	if not alignment2:
		PyMem_Free(alignment2)
		raise MemoryError()

	cdef char* p1 = alignment1
	cdef char* p2 = alignment2

	i = m
	j = n

	# first, walk from the lower right corner to the
	# position where we found the maximum score

	cdef int errors = 0

	cdef int gaps_are_errors  # if gaps are currently errors, this is 1, otherwise it's 0
	gaps_are_errors = 0 if (flags & STOP_WITHIN_SEQ2) else 1
	if i == best_i:  # we are in the last row
		while j > best_j:
			p1[0] = GAPCHAR
			j -= 1
			p2[0] = s2[j]
			p1 += 1
			p2 += 1
			errors += gaps_are_errors
	else:  # we are in the last column
		gaps_are_errors = 0 if (flags & STOP_WITHIN_SEQ1) else 1
		while i > best_i:
			i -= 1
			p1[0] = s1[i]
			p2[0] = GAPCHAR
			p1 += 1
			p2 += 1
			errors += gaps_are_errors

	assert i == best_i and j == best_j

	# the actual backtracing
	# The alignments are constructed in reverse
	# and this is undone afterwards.
	cdef int direction
	while i > 0 and j > 0:
		direction = columns[i,j].backtrace
		if direction == DIAG:
			i -= 1
			j -= 1
			if s1[i] != s2[j]:
				errors += 1
			p1[0] = s1[i]
			p2[0] = s2[j]
			p1 += 1
			p2 += 1
		elif direction == LEFT:
			errors += 1
			p1[0] = GAPCHAR
			j -= 1
			p2[0] = s2[j]
			p1 += 1
			p2 += 1
		elif direction == UP:
			i -= 1
			p1[0] = s1[i]
			p2[0] = GAPCHAR
			errors += 1
			p1 += 1
			p2 += 1
		else:
			assert False, 'DP table corrupt'

	cdef int start1 = i if (flags & START_WITHIN_SEQ1) else 0
	cdef int start2 = j if (flags & START_WITHIN_SEQ2) else 0

	errors += (i - start1) + (j - start2)

	while j > 0:
		p1[0] = GAPCHAR
		j -= 1
		p2[0] = s2[j]
		p1 += 1
		p2 += 1
	while i > 0:
		i -= 1
		p1[0] = s1[i]
		p2[0] = GAPCHAR
		p1 += 1
		p2 += 1
	assert i == 0 and j == 0

	align1 = alignment1[:(p1-alignment1)]
	align2 = alignment2[:(p2-alignment2)]

	align1 = align1[::-1]
	align2 = align2[::-1]

	PyMem_Free(alignment1)
	PyMem_Free(alignment2)

	return (align1, align2, start1, best_i, start2, best_j, errors)


def byte_frequencies(bytes s):
	"""Faster replacement for collections.Counter(s) for the case when s is a bytes object.

	Speed advantage depends on the length of the bytes object. When the length is less
	than 10, speedup is at least 2x. For length 100, speedup is 14x. For length 1000, speedup is
	approx 100x.
	"""
	cdef int[256] frequencies
	cdef int i
	cdef unsigned char c
	for i in range(256):
		frequencies[i] = 0
	for c in s:
		frequencies[c] += 1
	counter = Counter()
	for i in range(256):
		if frequencies[i] > 0:
			counter[i] = frequencies[i]
	return counter


def expected_errors(str qualities, int base=33):
	cdef int i, q
	cdef bytes quals = qualities.encode()
	cdef char* cq = quals
	cdef double e = 0.0
	for i in range(len(qualities)):
		q = cq[i] - base
		e += 10 ** (-q / 10)
	return e


def hamming_distance(unicode s, unicode t):
	"""
	Compute hamming distance between two strings. The two strings must have the
	same length.

	Return the number of differences between the strings.
	"""
	cdef Py_ssize_t m = len(s)
	cdef Py_ssize_t n = len(t)
	if m != n:
		raise IndexError("sequences must have the same length")
	cdef Py_ssize_t e = 0
	cdef Py_ssize_t i
	for i in range(m):
		if s[i] != t[i]:
			e += 1
	return e


# Lookup table that maps nucleotides to their 2-bit representation
# and everything else to 255.
cdef bytearray _nt_trans = bytearray([255]*256)
for frm, to in zip(b'ACGTacgt', b'\x00\x01\x02\x03\x00\x01\x02\x03'):
	_nt_trans[frm] = to


# Lookup table that maps 6-bit encoded codons to amino acids
def _make_codon_array(stop_aa='*'):
	triples = bytearray([ord(stop_aa)]*64)
	for codon, aa in GENETIC_CODE.items():
		b = codon.encode().translate(_nt_trans)
		index = b[0] * 16 + b[1] * 4 + b[2]
		triples[index] = ord(aa)
	return triples

cdef bytearray _codon_array = _make_codon_array()


def nt_to_aa(s: str):
	"""
	Translate a sequence of nucleotides to a sequence of amino acids,
	using the genetic code.

	>>> nt_to_aa('AAA')
	'K'
	>>> nt_to_aa('AAATGATGG)
	'K*W'
	"""
	cdef int i = 0
	cdef int j = 0
	cdef int v = 0
	cdef unsigned char nt0, nt1, nt2
	cdef char* nt_trans_ptr = <bytes>_nt_trans
	cdef char* codon_array_ptr = <bytes>_codon_array
	cdef bytes s_bytes = s.encode()
	cdef char* b = s_bytes
	cdef bytearray result = bytearray([0]*((len(s)+2)//3))
	cdef char* c = result
	cdef int n = len(b)
	for i in range(0, n-2, 3):
		v = 0
		nt0 = nt_trans_ptr[b[i]]
		nt1 = nt_trans_ptr[b[i+1]]
		nt2 = nt_trans_ptr[b[i+2]]
		if nt0 > 3 or nt1 > 3 or nt2 > 3:
			raise ValueError("Encountered non-nucleotide character in codon {!r}".format(s[i:i+3]))
		v = nt0 * 16 + nt1 * 4 + nt2
		c[j] = codon_array_ptr[v]
		j += 1
	if i < n:
		c[j] = '*'
	return result.decode()



def frequency_median(frequencies):
	"""
	Given a dictionary of frequencies, return the median.
	If the total no. of values is odd, the left of both
	middle values is returned.
	"""
	m = 0  # partial sum
	middle = (1 + sum(frequencies.values())) // 2
	for length in sorted(frequencies):
		m += frequencies[length]
		if m >= middle:
			return length
	# never reached
	assert False


def frequency_n50(lengths, genome_size=None):
	"""
	Return N50 or NG50 value given a Counter of lengths.

	If the genome_size is not given, it is set to the total length. The resulting
	value is the N50. If it is given, the resulting value is the NG50.

	If genome_size is greater than two times the number of sequences, None is returned.
	"""
	if genome_size is None:
		genome_size = sum(length * count for length, count in lengths.items())
	running_total = 0
	for length in sorted(lengths, reverse=True):
		running_total += length * lengths[length]
		if running_total >= genome_size * 0.5:
			return length
	return None


def n50(lengths, genome_size=None):
	"""
	Return N50 or NG50 value given a list of lengths.

	If the genome_size is not given, it is set to sum(lengths). The resulting
	value is the N50. If it is given, the resulting value is the NG50.

	If genome_size is greater than 2 * sum(lengths), None is returned.
	"""
	if genome_size is None:
		genome_size = sum(lengths)
	lengths = sorted(lengths, reverse=True)
	running_total = 0
	for length in lengths:
		running_total += length
		if running_total >= genome_size * 0.5:
			return length
	return None



#!/usr/bin/env python3
"""
Remove vector sequence

Limitations:
- Only one vector sequence supported.
- Only the first and last 500 bp of the vector sequence are actually searched for.
- If sequence is circular, vector might not be found.
"""
import sys
from cutadapt import align
from ..dna import reverse_complement
from .. import FastaWriter, SequenceReader, FastaReader
from argparse import ArgumentParser

def add_arguments(parser):
	arg = parser.add_argument
	arg('vector', help='FASTA with vector sequence(s)')
	arg('reads', help='FASTA/FASTQ with read')


def uncircularize(s, error_rate=0.05):
	"""
	Given a sequence s that overlaps itself (a prefix equals a suffix), return
	the sequence without the redundant suffix. To avoid random matches, the
	redundant sequence is only removed if it has a length of at least
	100 characters or 10% of the length of s, whichever is smaller.

	>>> uncircularize('hellotherehallo', error_rate=0.2)
	'hellothere'
	"""
	flags = align.START_WITHIN_SEQ2 | align.STOP_WITHIN_SEQ1

	# We cannot just align the sequence to itself since that results in a not
	# very helpful alignment, so take a prefix.
	k = int(len(s) * max(0.25, error_rate)) + 1
	prefix = s[:-k]
	result = align.locate(prefix, s, max_error_rate=error_rate, flags=flags)
	if result is None:
		return s
	pstart, pstop, sstart, sstop, matches, errors = result
	if pstop - pstart > min(100, int(0.1 * len(s))) and errors / (pstop - pstart) <= error_rate:
		# found overlap
		return s[:sstart]
	else:
		return s


def main(args=None):
	if args is None:
		parser = HelpfulArgumentParser(description=__doc__)
		add_arguments(parser)
		args = parser.parse_args()

	vectors = list(FastaReader(args.vector))
	assert len(vectors) == 1, "Only FASTA files with exactly one vector sequence currently supported"

	vector = vectors[0].sequence.upper()
	flags = align.START_WITHIN_SEQ2 | align.STOP_WITHIN_SEQ2
	prefix = vector[:500]
	suffix = vector[-500:]

	writer = FastaWriter(sys.stdout, line_length=80)
	for read in SequenceReader(args.reads):
		print('Working on', len(read.sequence), 'bp read', repr(read.name), file=sys.stderr)
		seq = read.sequence.upper()
		uncirc = uncircularize(seq)
		if len(uncirc) < len(seq):
			print('    uncircularized length:', len(uncirc), file=sys.stderr)
			circular = True
			seq = uncirc
		else:
			print('    not circular', file=sys.stderr)
			circular = False
		# The vector is either within the sequence:
		# ---XXXXXX---
		# or (due to circularity) a suffix and prefix:
		# XX------XXXX

		# So by appending the sequence to itself, we make certain that a full
		# copy of the vector sequence appears:
		# XX------XXXXXX---...

		if circular:
			s = seq + seq  # TODO 500 would be enough as long as we're only checking prefix and suffix
		else:
			s = seq
		for revcomp in False, True:
			if revcomp:
				s = reverse_complement(s)

			# Search for prefix of the vector
			result = align.locate(prefix, s, max_error_rate=0.1, flags=flags)
			if result is None:
				continue
			vstart, vstop, start, _, _, errors = result
			prefix_erate = errors / (vstop - vstart)
			assert prefix_erate <= 0.1
			print('    Prefix match found with {:.2%} errors'.format(prefix_erate), file=sys.stderr)

			# Search for suffix of the vector, but not before the prefix
			vstart, vstop, _, stop, _, errors = align.locate(suffix, s[start:], max_error_rate=0.1, flags=flags)
			suffix_erate = errors / (vstop - vstart)
			if suffix_erate > 0.1:
				print('    Suffix error rate too large.', file=sys.stderr)
				continue
			print('    Suffix match found with {:.2%} errors'.format(suffix_erate), file=sys.stderr)

			if circular:
				# We can output a single sequence.
				# Stop coordinate is relative to s[start:], so the actual
				# start is at start + stop.
				seq = s[start+stop:start+len(seq)]
				print('    Read trimmed. Length changed from', len(read.sequence), 'to', len(seq), file=sys.stderr)
				writer.write(read.name, seq)
			else:
				# Need to output two separate sequences here.
				seq1 = s[0:start]
				seq2 = s[stop:]
				print('    Read split into two. Length changed from', len(read.sequence), 'to', len(seq1), '+', len(seq2), file=sys.stderr)
				# Numbering is switched on purpose since seq1 actually
				# follows seq2 on the reference.
				writer.write(read.name + '-1', seq2)
				writer.write(read.name + '-2', seq1)
			break


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
"""
Simulate reads that optionally contain adapter sequences.
"""
import sys
from random import randint, random, seed

from sqt import HelpfulArgumentParser
from sqt.io.fasta import IndexedFasta

__author__ = "Marcel Martin"

LENGTH = 100

#adapter = 'GCCTAACTTCTTAGACTGCCTTAAGGACGT'
#adapter_prob = 0.5


def main():
	parser = HelpfulArgumentParser(description=__doc__)
	arg = parser.add_argument
	arg("--length", "-l", type=int, default=100)
	arg("--seed", type=int, default=None, help="seed for random number generator")
	#arg("--minimum-length", "-m", type=int, default=100)
	#arg("--maximum-length", "-M", type=int, default=100)
	arg("--adapter", "-a", default=None,
		help="Add an adapter")
	arg("--probability", "-p", default=0.5,
		help="Fraction of reads (approximate) that should contain the adapter")
	arg("n", type=int, help="Number of reads")
	arg("fasta", #nargs='?',
		help="FASTA file from which reads are sampled. Only the first chromosome (entry) is used.")
	arg("chromosome", help="chromosome that is picked from the FASTA file")

	args = parser.parse_args()
	if args.seed is not None:
		seed(args.seed)
	adapter = args.adapter
	adapter_prob = args.probability
	length = args.length
	fasta = IndexedFasta(args.fasta)
	chrom = fasta.get(args.chromosome)
	i = 0
	while i < args.n:
		start = randint(0, len(chrom) - 100)
		seq = chrom[start:start+length]
		if b'N' in seq:
			continue
		i += 1
		seq = seq.decode('ascii')
		if adapter is not None and random() < adapter_prob:
			pos = randint(0, length-1)
			seq = seq[:pos] + adapter + seq[pos:]
			seq = seq[:length]
			extra = ' adapterpos={}'.format(pos+1)
		else:
			extra = ''
		print('>r{} {}:{}-{}{}'.format(i, args.chromosome, start+1, start+length, extra))
		print(seq)


	#ALPHABET = 'ACGT'
	#for i in range(args.n):
		#l = randint(args.minimum_length, args.maximum_length)
		#seq = ''.join(choice('ACGT') for _ in range(l))
		#print(">r{0}\n{1}".format(i+1, seq))


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
"""
Print q-gram (also called k-mer) frequencies in a FASTA or FASTQ file.

The result is a list of q-grams and their counts, sorted by counts from
least to most frequent.
"""
import sys
from collections import Counter

from sqt import HelpfulArgumentParser
from sqt import SequenceReader

__author__ = "Marcel Martin"

def q_grams(s, q):
	"""yield all q-grams in s"""
	for i in range(len(s) - q):
		yield s[i:i+q]


def add_arguments(parser):
	arg = parser.add_argument
	arg("-q", default=4,
		help="length of the q-grams (also called k-mers) (default: %(default)s)")
	arg("path", metavar='FASTA/FASTQ', help="input FASTA or FASTQ file")


def main(args=None):
	if args is None:
		parser = HelpfulArgumentParser(description=__doc__)
		add_arguments(parser)
		args = parser.parse_args()

	counts = Counter()
	with SequenceReader(args.path) as reader:
		for record in reader:
			counts.update(q_grams(record.sequence, args.q))
	for elem, count in counts.most_common()[::-1]:
		print(elem, count)


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
"""
Read in reads in FASTA format, insert an adapter sequence at a random
position, shorten the read to its original length, and write it out.

An annotation is added to the read description that describes the
adapter position, such as "adapterpos=17". The position is 1-based.
"""
import sys
from random import randint, random, seed, choice

from sqt import HelpfulArgumentParser
from sqt.io.fasta import FastaReader

__author__ = "Marcel Martin"

#adapter = 'GCCTAACTTCTTAGACTGCCTTAAGGACGT'

def mutate_sequence(seq, rate=0.1, alphabet='ACGT', indels=False, indel_rate=0.1):
	"""
	If indels is True, the mutation_rate is equally split up between insertions,
	deletions and substitutions.
	"""
	mutated = []
	charpos = dict( (c,i) for (i,c) in enumerate(alphabet))
	for c in seq:
		c = c.upper()
		r = random()
		if indels and r < 0.5 * rate * indel_rate:
			# insertion
			mutated.append(choice(alphabet))
			mutated.append(c)
		elif indels and r < rate * indel_rate:
			# deletion
			pass
		elif r < rate:
			# mutate base
			d = alphabet[(charpos[c] + randint(1, len(alphabet)-1)) % len(alphabet)]
			assert d != c
			mutated.append(d)
		else:
			# no change
			mutated.append(c)
	return ''.join(mutated)


def main():
	parser = HelpfulArgumentParser(description=__doc__)
	arg = parser.add_argument
	arg("--seed", type=int, default=None, help="seed for random number generator")
	arg("--erate", type=float, default=None,
		help="error rate for simulated errors, no indels (default: no errors)")
	arg("--adapter", "-a", default='GCCTAACTTCTTAGACTGCCTTAAGGACGT',
		help="Adapter sequence")
	arg("--probability", "-p", default=0.5,
		help="Fraction of reads (approximate) that should contain the adapter")
	arg("fasta", help="Input FASTA file. Use '-' for standard input.")

	args = parser.parse_args()
	if args.seed is not None:
		seed(args.seed)
	adapter = args.adapter
	adapter_prob = args.probability
	with FastaReader(args.fasta) as reader:
		for read in reader:
			if random() < adapter_prob:
				if args.erate is not None:
					a = mutate_sequence(adapter, args.erate, indels=False)
				else:
					a = adapter
				l = len(read)
				pos = randint(0, l-1)
				seq = read.sequence
				seq = seq[:pos] + a + seq[pos:]
				read.sequence = seq[:l]
				read.name = read.name + ' adapterpos={}'.format(pos+1)
			print('>{}\n{}'.format(read.name, read.sequence))



if __name__ == '__main__':
	main()



#!/usr/bin/env python3
"""
Print the number of characters in a string.
"""
from sqt import HelpfulArgumentParser

__author__ = "Marcel Martin"


def add_arguments(parser):
	arg = parser.add_argument
	arg('string', help='The string')


def main(args=None):
	if args is None:
		parser = HelpfulArgumentParser(description=__doc__)
		add_arguments(parser)
		args = parser.parse_args()

	print(len(args.string))


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
"""
Search for a IUPAC string in the sequences of a FASTA file.

Prints matching entries in the fasta file to standard output.
If <FASTA> is not provided, read from standard input.

If output is a terminal, the first occurrence of the pattern
in each sequence is highlighted.
"""
import sys
import re
from sqt import HelpfulArgumentParser
from sqt.io.fasta import FastaReader
from sqt.io.xopen import xopen
from sqt.ansicolor import red, lightred

__author__ = "Marcel Martin, Tobias Marschall"


def add_arguments(parser):
	arg = parser.add_argument
	arg("-d", "--description", action="store_true", default=False,
		help="Search the description/comment fields of the FASTA file instead "
		"of the sequences. If given, the pattern is interpreted as a regular "
		"expression, not as a IUPAC pattern. (default: %(default)s)")
	arg("pattern")
	arg("fasta", nargs='?', const=None)


def iupac_to_regex(iupac):
	"""
	Converts a IUPAC string with wildcards to a regular expression.
	"""
	wildcards = {
		"R": "AG",
		"Y": "CT",
		"S": "CG",
		"W": "AT",
		"K": "GT",
		"M": "AC",
		"B": "CGT",
		"D": "AGT",
		"H": "ACT",
		"V": "ACG",
		"N": "ACGT",
		"X": "ACGT" }
	regex = ""
	for c in iupac.upper():
		if c in "ACGT":
			regex += c
		elif c in wildcards:
			regex += "[" + wildcards[c] + "]"
		else:
			raise ValueError("don't know how to handle character %s" % c)
	return regex



def main(args=None):
	if args is None:
		parser = HelpfulArgumentParser(description=__doc__)
		add_arguments(parser)
		args = parser.parse_args()

	if args.fasta is None:
		infile = sys.stdin
	else:
		infile = xopen(args.fasta)

	# whether to use color in output
	color = sys.stdout.isatty()

	if args.description:
		reg = args.pattern
		color = False
	else:
		reg = iupac_to_regex(args.pattern.upper())
	regex = re.compile(reg)
	for record in FastaReader(infile):
		seq = record.sequence
		#print(desc, seq)
		if args.description:
			match = regex.search(record.name)
		else:
			match = regex.search(record.sequence.upper())
		if match:
			print('>', record.name, sep='')
			if color:
				print(seq[0:match.start()] + lightred(seq[match.start():match.end()]) + seq[match.end():])
			else:
				print(seq)
			match = regex.search(seq.upper())


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
"""
Repair a BAM file that (incorrectly) contains phred64-encoded qualities.
BAM files must encode their quality values in phred33.

Output is written in SAM format to standard output.
"""
from sqt import HelpfulArgumentParser
from pysam import Samfile

__author__ = "Marcel Martin"

PHRED_64_TO_33_TRANS = bytes.maketrans(bytes(range(64, 64+100)), bytes(range(33, 33+100)))

def get_argument_parser():
	parser = HelpfulArgumentParser(description=__doc__)
	add = parser.add_argument
	add('bam', help='Input BAM file')
	return parser


def main():
	parser = get_argument_parser()
	args = parser.parse_args()

	with Samfile(args.bam) as infile:
		with Samfile('-', 'wh', template=infile) as outfile:
			for record in infile:
				record.qual = record.qual.translate(PHRED_64_TO_33_TRANS)
				outfile.write(record)


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
"""
Guess quality encoding of one or more FASTA files.
"""
import sys
import os
import subprocess
from collections import Counter
from sqt.io.fasta import FastqReader, guess_quality_base
from sqt import HelpfulArgumentParser

__author__ = "Marcel Martin"


def get_argument_parser():
	parser = HelpfulArgumentParser(description=__doc__)
	add = parser.add_argument
	add('--verbose', '-v', default=False, action='store_true',
		help='Print histogram of found characters')
	add('--limit', '-n', default=10000, type=int,
		help='Inspect the first LIMIT records in the FASTQ file (default: %(default)s)')
	add('fastq', nargs='+', metavar='FASTQ',
		help='Input FASTQ files (may be gzipped).')
	return parser


def main():
	parser = get_argument_parser()
	args = parser.parse_args()

	for path in args.fastq:
		if args.verbose:
			print('## File:', path)
		else:
			print(path, end='')
		freqs, guess = guess_quality_base(path)
		if args.verbose:
			print()
			print('character ASCII frequency')
			for c in sorted(freqs):
				print("{} {:3} {:7}".format(chr(c), c, freqs[c]))
			print()
		else:
			print(' is ', end='')
		guess = { 33: 'phred33', 64: 'phred64', None: 'unknown'}[guess]
		if args.verbose:
			print("Quality value range assuming phred33: {}..{}".format(min(freqs) - 33, max(freqs) - 33))
			print("Quality value range assuming phred64: {}..{}".format(min(freqs) - 64, max(freqs) - 64))
			print("This is probably", guess)
		else:
			print(guess)


if __name__ == '__main__':
	main()



#!/usr/bin/env python3
"""
Perform set operation on two SAM/BAM files.

The output BAM file will have the same header as file A.

WARNING: Implementation is neither very fast nor memory efficient.

Possible operations:
  union:        Output union of A and B, abort with error if
                different lines for same read are encountered.
  intersection: Output intersection of A and B, abort with error if
                different lines for same read are encountered.
  setminus:     Outputs all read in A that are not in B.
  symdiff:      Output all reads in A or B but not in both.
"""
from sqt import HelpfulArgumentParser
import sys
from pysam import Samfile

__author__ = "Tobias Marschall"

def add_arguments(parser):
	arg = parser.add_argument
	arg("-s", action="store_true", dest="sam_output", default=False,
		help="Output SAM file instead of BAM file")
	arg("-U", action="store_true", dest="exclude_unmapped_A", default=False,
		help="Exclude unmapped reads from file A")
	arg("-V", action="store_true", dest="exclude_unmapped_B", default=False,
		help="Exclude unmapped reads from file B")
	arg("-r", action="store_true", dest="remove_name_suffix", default=False,
		help="Remove trailing \"/*\" from read names. Useful if one mapper appends \"/1\" and another does not.")
	arg('bampath1', help='First BAM or SAM file')
	arg('operation', choices=('union','intersection','setminus','symdiff'))
	arg('bampath2', help='Second BAM or SAM file')
	arg('outputpath', nargs='?',
		help='Output BAM or SAM file. If omitted, only print the number of reads '
		'that would be written.')


def SamOrBam(name, mode='r'):
	if name.endswith('.bam'):
		mode += 'b'
	return Samfile(name, mode)


def remove_suffix(s):
	i = s.rfind('/')
	if i == -1: return s
	else: return s[:i]


def nop(s):
	return s


def dict_of_reads(reads, exclude_unmapped, rename):
	d = dict()
	for read in reads:
		if exclude_unmapped and read.is_unmapped: continue
		name = rename(read.qname)
		if d.has_key(name):
			raise Exception("Duplicate read in input file (%s)"%name)
		d[name] = read
	return d


def union(A, B,outfile, exclude_unmapped_A, exclude_unmapped_B, rename):
	readsB = dict_of_reads(B, exclude_unmapped_B, rename)
	readnamesA = set()
	for read in A:
		if exclude_unmapped_A and read.is_unmapped: continue
		name = rename(read.qname)
		if name in readnamesA:
			raise Error("Duplicate read in input file (%s)"%name)
		readnamesA.add(name)
		if readsB.has_key(name):
			if read.compare(readsB[name]) != 0:
				print('Content mismatch for read %s:'%name, file=sys.stderr)
				print('File A:',read, file=sys.stderr)
				print('File B:',readsB[name], file=sys.stderr)
				sys.exit(1)
			readsB.pop(name)
		outfile.write(read)
	for read in readsB.itervalues():
		outfile.write(read)


def intersection(A,B,outfile,exclude_unmapped_A,exclude_unmapped_B,rename):
	readsB = dict_of_reads(B, exclude_unmapped_B, rename)
	readnamesA = set()
	for read in A:
		if exclude_unmapped_A and read.is_unmapped: continue
		name = rename(read.qname)
		if name in readnamesA:
			raise Error("Duplicate read in input file (%s)"%name)
		readnamesA.add(name)
		if readsB.has_key(name):
			if read.compare(readsB[name])!=0:
				print('Content mismatch for read %s:'%name, file=sys.stderr)
				print('File A:',read, file=sys.stderr)
				print('File B:',readsB[name], file=sys.stderr)
				sys.exit(1)
			outfile.write(read)


def setminus(A,B,outfile,exclude_unmapped_A,exclude_unmapped_B,rename):
	if exclude_unmapped_B: readnamesB = set((rename(read.qname) for read in B if not read.is_unmapped))
	else: readnamesB = set((rename(read.qname) for read in B))
	for read in A:
		if exclude_unmapped_A and read.is_unmapped: continue
		if not rename(read.qname) in readnamesB:
			outfile.write(read)


def symdiff(A,B,outfile,exclude_unmapped_A,exclude_unmapped_B,rename):
	if exclude_unmapped_B: readsB = dict(((rename(read.qname),read) for read in B if not read.is_unmapped))
	else: readsB = dict(((rename(read.qname),read) for read in B))
	for read in A:
		if exclude_unmapped_A and read.is_unmapped: continue
		name = rename(read.qname)
		if readsB.has_key(name):
			readsB.pop(name)
		else:
			outfile.write(read)
	for read in readsB.itervalues():
		outfile.write(read)


class Counter:
	count = 0
	def write(self, x):
		self.count += 1


def main(args=None):
	if args is None:
		parser = HelpfulArgumentParser(description=__doc__)
		add_arguments(parser)
		args = parser.parse_args()
	operation = args.operation
	A = SamOrBam(args.bampath1)
	B = SamOrBam(args.bampath2)
	if args.outputpath:
		outfile = Samfile(args[3], 'wh' if args.sam_output else 'wb', template=A)
	else:
		outfile = Counter()
	globals()[operation](A, B, outfile, args.exclude_unmapped_A, args.exclude_unmapped_B, remove_suffix if args.remove_name_suffix else nop)
	if args.outputpath is None:
		print(outfile.count)


if __name__ == '__main__':
	main()



from glob import glob
import os
import sys

from setuptools import setup, Extension
from distutils.version import LooseVersion
from distutils.command.sdist import sdist as _sdist
from distutils.command.build_ext import build_ext as _build_ext

import versioneer

MIN_CYTHON_VERSION = '0.17'

if sys.version_info < (3, 3):
	sys.stdout.write("At least Python 3.3 is required.\n")
	sys.exit(1)


def no_cythonize(extensions, **_ignore):
	"""
	Change file extensions from .pyx to .c or .cpp.

	Copied from Cython documentation
	"""
	for extension in extensions:
		sources = []
		for sfile in extension.sources:
			path, ext = os.path.splitext(sfile)
			if ext in ('.pyx', '.py'):
				if extension.language == 'c++':
					ext = '.cpp'
				else:
					ext = '.c'
				sfile = path + ext
			sources.append(sfile)
		extension.sources[:] = sources


def check_cython_version():
	"""exit if Cython not found or out of date"""
	try:
		from Cython import __version__ as cyversion
	except ImportError:
		sys.stdout.write(
			"ERROR: Cython is not installed. Install at least Cython version " +
			str(MIN_CYTHON_VERSION) + " to continue.\n")
		sys.exit(1)
	if LooseVersion(cyversion) < LooseVersion(MIN_CYTHON_VERSION):
		sys.stdout.write(
			"ERROR: Your Cython is at version '" + str(cyversion) +
			"', but at least version " + str(MIN_CYTHON_VERSION) + " is required.\n")
		sys.exit(1)


cmdclass = versioneer.get_cmdclass()
versioneer_build_ext = cmdclass.get('build_ext', _build_ext)
versioneer_sdist = cmdclass.get('sdist', _sdist)


class build_ext(versioneer_build_ext):
	def run(self):
		# If we encounter a PKG-INFO file, then this is likely a .tar.gz/.zip
		# file retrieved from PyPI that already includes the pre-cythonized
		# extension modules, and then we do not need to run cythonize().
		if os.path.exists('PKG-INFO'):
			no_cythonize(extensions)
		else:
			# Otherwise, this is a 'developer copy' of the code, and then the
			# only sensible thing is to require Cython to be installed.
			check_cython_version()
			from Cython.Build import cythonize
			self.extensions = cythonize(self.extensions)
		versioneer_build_ext.run(self)


class sdist(versioneer_sdist):
	def run(self):
		# Make sure the compiled Cython files in the distribution are up-to-date
		from Cython.Build import cythonize
		check_cython_version()
		cythonize(extensions)
		versioneer_sdist.run(self)


cmdclass['build_ext'] = build_ext
cmdclass['sdist'] = sdist

extensions = [
	Extension('sqt._helpers', sources=['sqt/_helpers.pyx']),
]


setup(
	name = 'sqt',
	version = versioneer.get_version(),
	author = 'Marcel Martin',
	author_email = 'marcel.martin@scilifelab.se',
	url = 'https://bitbucket.org/marcelm/sqt',
	description = 'Command-line tools for the analysis of high-throughput sequencing data',
	license = 'MIT',
	cmdclass = cmdclass,
	packages = [ 'sqt', 'sqt.io', 'sqt.commands' ],
	entry_points = {'console_scripts': [
		'sqt = sqt.__main__:main',
		#'sqt-addadapt = sqt.commands.addadapt:main',
		#'sqt-bam2fastq = sqt.commands.bam2fastq:main',
		#'sqt-bamstats = sqt.commands.bamstats:main',
		#'sqt-checkfastqpe = sqt.commands.checkfastqpe:main',
		#'sqt-checkvcfref = sqt.commands.checkvcfref:main',
		#'sqt-compare-sequences = sqt.commands.compare_sequences:main',
		#'sqt-coverage = sqt.commands.coverage:main',
		#'sqt-fastaextract = sqt.commands.fastaextract:main',
		#'sqt-fastamutate = sqt.commands.fastamutate:main',
		#'sqt-fastastats = sqt.commands.fastastats:main',
		#'sqt-fastxmod = sqt.commands.fastxmod:main',
		#'sqt-fixbam64 = sqt.commands.fixbam64:main',
		#'sqt-globalalign = sqt.commands.globalalign:main',
		#'sqt-histogram = sqt.commands.histogram:main',
		#'sqt-qualityguess = sqt.commands.qualityguess:main',
		#'sqt-readcov = sqt.commands.readcov:main',
		#'sqt-samfixn = sqt.commands.samfixn:main',
		#'sqt-simreads = sqt.commands.simreads:main',
		#'sqt-translate = sqt.commands.translate:main',
	]},
	install_requires = [
		'pysam!=0.9.0',
		'cutadapt',
		'matplotlib',
		'seaborn',
	],
	ext_modules = extensions,
	test_suite = 'nose.collector',
	classifiers = [
		"Development Status :: 3 - Alpha",
		#Development Status :: 4 - Beta
		#Development Status :: 5 - Production/Stable
		"Environment :: Console",
		"Intended Audience :: Science/Research",
		"License :: OSI Approved :: MIT License",
		"Natural Language :: English",
		"Programming Language :: Python :: 3",
		"Topic :: Scientific/Engineering :: Bio-Informatics"
	]
)



import os
import re

def available_cpu_count():
	"""
	Number of available virtual or physical CPUs on this system.

	Adapted from http://stackoverflow.com/a/1006301/715090
	"""
	# cpuset may restrict the number of available processors
	cpus = 1
	try:
		m = re.search(r'(?m)^Cpus_allowed:\s*(.*)$',
						open('/proc/self/status').read())
		if m:
			res = bin(int(m.group(1).replace(',', ''), 16)).count('1')
			if res > 0:
				cpus = res
	except IOError:
		pass

	try:
		import multiprocessing
		cpus = min(cpus, multiprocessing.cpu_count())
	except (ImportError, NotImplementedError):
		pass
	return cpus



#!/usr/bin/env python3
"""
Print a report for individual reads in a SAM/BAM file.
"""
__author__ = 'Marcel Martin'

import sys
from collections import Counter, namedtuple, defaultdict
from itertools import islice
from contextlib import ExitStack
from pysam import Samfile
from sqt import HelpfulArgumentParser, Cigar, cigar
from sqt.region import Region

from .bamstats import AlignedRead, print_coverage_report

def add_arguments(parser):
	arg = parser.add_argument
	arg('--minimum-length', '-m', type=int, default=1,
		help='Minimum read length. Ignore reads that are shorter. Default: %(default)s')
	arg('--quality', '-q', type=int, default=0,
		help='Minimum mapping quality (default: %(default)s')
	arg('--minimum-cover-fraction', metavar='FRACTION', type=float, default=0.01,
		help='Alignment must cover at least FRACTION of the read to appear in the cover report. (%(default)s)')
	arg("bam", metavar="SAM/BAM", help="Name of a SAM or BAM file")
	arg("region", help="Region")


def main(args=None):
	if args is None:
		parser = HelpfulArgumentParser(description=__doc__)
		add_arguments(parser)
		args = parser.parse_args()
	region = Region(args.region)
	n_records = 0
	seen_reads = set()
	with Samfile(args.bam) as sf:
		for record in sf.fetch(region.reference, region.start, region.stop):
			if record.query_length < args.minimum_length:
				continue
			n_records += 1
			if record.is_unmapped:
				unmapped += 1
				unmapped_bases += len(record.seq)
				continue
			if record.mapq < args.quality:
				continue
			assert record.cigar is not None

			if not record.query_name in seen_reads:
				aligned_read = AlignedRead(record, sf.getrname(record.tid))
				print_coverage_report(aligned_read, minimum_cover_fraction=args.minimum_cover_fraction)
				seen_reads.add(record.query_name)


if __name__ == '__main__':
	main()
